/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: cpu_clk.c
 *
 * Function: the CPU Speed adjustment
 *
 * Part of:  DIY Calculator loader application
 *
 *
 * Changelog
 *
 *  0.1   2.Nov.2006   JH   new
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>

#include <string.h>          /* strchr() */
#include <stdlib.h>          /* atoi() */

#ifndef WIN32
#include <unistd.h>          /* usleep() */
#endif

#include "diy_load.h"
#include "serial.h"
#include "cmd.h"
#include "cpu_clk.h"

/******************************  Defines  *********************************/
enum UNITS {HZ, KHZ, MHZ};
enum BUTTONS {OK_BUTTON, APPLY_BUTTON, CLOSE_BUTTON};

/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/**************************  Private Functions  ***************************/


/************************  Function Prototypes  ***************************/


/******************************  Globals  *********************************/
static GtkWidget *window;    /* the CPU speed adjustment window           */
static GtkWidget *cpu_clk_entry, *cpu_clk_label;
static GtkObject *cpu_clk_adjust;

static unsigned char offline_divider=0;/* fake divider in offline mode    */

extern int debug;
extern int offline;
extern int do_timer_cb;


/* cpu_clk_window_destroy */
/**************************************************************************/
void cpu_clk_window_destroy (     /* destroy the CPU speed adj. window    */
  GtkWidget *widget,              /* widget which triggered the callback  */
  gpointer data)                  /* user data                            */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the CPU Speed Adjustment window                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the CPU speed adjustment window.                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* cpu_clk_button_cb */
/**************************************************************************/
void cpu_clk_button_cb (     /* OK/Apply/Close button callback function   */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : window, offline, offline_divider                           *
* called from : on clicking any of the OK/Apply/Close buttons              *
* calls       : cpu_clk_window_destroy(), write_port()                     *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* OK/Apply/Close button callback function                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  unsigned char value;
  char cmd[20];
  int button;
  static const char *button_names[]={"OK", "Apply", "Close"};

  button = (int) data;
  DEBUG (("cpu_clk_button_cb() %s button\n", button_names[button]));

  if (button < 2)
  {
    /* send the new value to the DIY Calculator */
    value = (unsigned char)
      gtk_adjustment_get_value (GTK_ADJUSTMENT(cpu_clk_adjust));
    if (offline)
      offline_divider = (unsigned char)(~value);
    else
    {
      sprintf (cmd, "mw %2.2x f036 1\n", (unsigned char)(~value));
      write_port (cmd, strlen(cmd));
    }
  }
  if ((button&1) == 0)
  {
    /* close the window */
    cpu_clk_window_destroy (window, NULL);
  }
}

/* cpu_clk_slider_cb */
/**************************************************************************/
void cpu_clk_slider_cb (     /* slider callback function                  */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : cpu_clk_entry, cpu_clk_label                               *
* called from : on moving the horizontal slider                            *
* calls       : divider_to_freq()                                          *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Slider callback function                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  unsigned char value;
  unsigned int freq;
  char buf[BUFLEN], *p;

  DEBUG (("cpu_clk_slider_cb(): "));

  value = (unsigned char) gtk_adjustment_get_value (GTK_ADJUSTMENT (widget));
  DEBUG (("value=%d, ", value));
  freq = divider_to_freq (~value, buf);
  DEBUG (("freq=%u\n", freq));
  p = strchr (buf, ' ');
  *p = '\0';
  p ++;
  gtk_entry_set_text (GTK_ENTRY(cpu_clk_entry), buf);
  gtk_label_set_text (GTK_LABEL(cpu_clk_label), p);
}

/* do_cpu_clk */
/**************************************************************************/
GtkWidget *do_cpu_clk (      /* the CPU Speed Adjustment                  */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : window, offline, cpu_clk_entry, cpu_clk_label, do_timer_cb *
*               offline_divider, cpu_clk_adjust                            *
* called from : set_cpu_clk()                                              *
* calls       : write_port(), read_line(), atox()                          *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Create the CPU speed adjustment window.                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* pointer to the CPU speed adjustment window object                        *
***************************************************************************/
{
  GtkWidget *overall_box, *vbox2, *vbox3, *slider_box;
  GtkWidget *button_box, *minmax_box, *hbox4;
  GtkWidget *slider, *button, *label;
  char cmd[]="md f036 1\n";
  char buf[BUFLEN];
  unsigned char divider;

  DEBUG (("do_cpu_clk()\n"));
  if (window == NULL)
  {
    /**************  create the "CPU Speed Adjust" window  **************/
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    /* install the delete_event handler that immediately exits GTK */
    g_signal_connect (G_OBJECT(window), "delete_event",
      G_CALLBACK (cpu_clk_window_destroy), NULL);
    /* set the border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (window), 10);
    /* set the window title */
    if (offline)
      gtk_window_set_title(GTK_WINDOW(window),"Set the CPU Clock - offline");
    else
      gtk_window_set_title (GTK_WINDOW (window), "Set the CPU Clock");
    /* when the main window will be closed, close also this one */
    gtk_window_set_destroy_with_parent (GTK_WINDOW(window), TRUE);
    /* set it on top of the main window */
    gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));

    /*******************   create the various widgets  **********************/
    overall_box = gtk_vbox_new (FALSE, 10);
    /* put the box into the window */
    gtk_container_add (GTK_CONTAINER (window), overall_box);

    /* a box for the min/max text, slider, entry box, and units label */
    slider_box = gtk_hbox_new (FALSE, 10);
    gtk_container_add (GTK_CONTAINER (overall_box), slider_box);

    /* at the left side we put the min/max test and the slider */
    vbox2 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (slider_box), vbox2);

    minmax_box = gtk_hbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (vbox2), minmax_box);

    label = gtk_label_new ("min");
    gtk_box_pack_start (GTK_BOX(minmax_box), label, FALSE, TRUE, 0);

    label = gtk_label_new ("max");
    gtk_box_pack_end (GTK_BOX(minmax_box), label, FALSE, TRUE, 0);

    cpu_clk_adjust = gtk_adjustment_new (0.0, 0.0, 255.0, 1.0, 1.0, 0.0);
    g_signal_connect (G_OBJECT (cpu_clk_adjust), "value_changed",
                        G_CALLBACK (cpu_clk_slider_cb), NULL);
    slider = gtk_hscale_new (GTK_ADJUSTMENT(cpu_clk_adjust));
    gtk_widget_set_size_request (GTK_WIDGET (slider), 200, -1);
    gtk_scale_set_draw_value (GTK_SCALE (slider), FALSE);
    gtk_container_add (GTK_CONTAINER (vbox2), slider);

    /* at the right side we put the entry box and the units label */
    vbox3 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (slider_box), vbox3);

    /* dummy place holder */
    label = gtk_label_new (" ");
    gtk_box_pack_start (GTK_BOX(vbox3), label, TRUE, TRUE, 0);

    /* the box for the entry box and the units label */
    hbox4 = gtk_hbox_new (FALSE, 4);
    gtk_container_add (GTK_CONTAINER (vbox3), hbox4);

    cpu_clk_entry = gtk_entry_new ();
    gtk_widget_set_size_request (GTK_WIDGET (cpu_clk_entry), 30, -1);
    gtk_entry_set_editable (GTK_ENTRY(cpu_clk_entry), FALSE);
    gtk_container_add (GTK_CONTAINER (hbox4), cpu_clk_entry);

    cpu_clk_label = gtk_label_new ("MHz");
    gtk_container_add (GTK_CONTAINER (hbox4), cpu_clk_label);

    /* dont let the slider or the entry grow when the value changes */  
    gtk_box_set_child_packing (GTK_BOX (slider_box),
      vbox2, FALSE, FALSE, 0, GTK_PACK_START);
    gtk_box_set_child_packing (GTK_BOX (slider_box),
      vbox3, FALSE, FALSE, 0, GTK_PACK_START);

    /* the buttons */
    button_box = gtk_hbox_new (TRUE, 20);
    gtk_container_add (GTK_CONTAINER (overall_box), button_box);

    button = gtk_button_new_with_label ("OK");
    g_signal_connect (G_OBJECT (button), "clicked",
      G_CALLBACK (cpu_clk_button_cb), (gpointer)OK_BUTTON);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    button = gtk_button_new_with_label ("Apply");
    g_signal_connect (G_OBJECT (button), "clicked",
      G_CALLBACK (cpu_clk_button_cb), (gpointer)APPLY_BUTTON);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    button = gtk_button_new_with_label ("Close");
    g_signal_connect (G_OBJECT (button), "clicked",
      G_CALLBACK (cpu_clk_button_cb), (gpointer)CLOSE_BUTTON);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_CENTER_ON_PARENT);

    gtk_adjustment_value_changed (GTK_ADJUSTMENT(cpu_clk_adjust));

    /* show the window */
    gtk_widget_show_all (window);
  }  /* have to create the CPU clock window */

  if (offline)
    divider = offline_divider;
  else
  {
    /* retrieve the current CPU divider */
    do_timer_cb = 0;
    write_port (cmd, strlen(cmd));
    usleep (200000);
  
    /* line with the "mw f036 1" cmd */
    read_line (buf, BUFLEN);
    DEBUG (("do_cpu_clk(): 1 buf=%s", buf));
  
    /* empty line */
    read_line (buf, BUFLEN);
    DEBUG (("do_cpu_clk(): 2 buf=%s", buf));
  
    /* empty line */
    read_line (buf, BUFLEN);
    DEBUG (("do_cpu_clk(): 3 buf=%s", buf));
  
    /* line "Memory Display ... */
    read_line (buf, BUFLEN);
    DEBUG (("do_cpu_clk(): 4 buf=%s", buf));
  
    /* line with interesting data */
    read_line (buf, BUFLEN);
    DEBUG (("do_cpu_clk(): 5 buf=%s", buf));
    buf[8] = '\0';
    do_timer_cb = 1;
    divider = atox (buf+6);
  }
  DEBUG (("do_cpu_clk(): divider=0x%2.2x\n", divider));
  gtk_adjustment_set_value (GTK_ADJUSTMENT (cpu_clk_adjust),
                            (unsigned char)~divider);
  return window;
}

/* divider_to_freq */
/**************************************************************************/
unsigned int divider_to_freq(/* compute the frequency from the divider    */
  unsigned char divider,     /* the frequency divider                     */
  char *s)                   /* representation as string                  */
/***************************************************************************
* globals     :                                                            *
* called from : cpu_clk_slider_cb(), fill_io_ports(), do_version_info()    *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Compute the CPU clock speed from the 8bit divider.                       *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* CPU frequency                                                            *
***************************************************************************/
{
  unsigned int freq;
  int segment, offset;
  int units;
  float display;
  char *unit_name[]={"Hz","kHz","MHz"};

//  DEBUG (("divider_to_freq(): divider=0x%2.2x\n", divider));
  if (divider == 0)
    divider ++;
  segment = divider >> 4;
  offset = divider & 0x0f;
  freq = 0;
  switch (segment)
  {
    case  0: freq = 0x000000 + (offset << 1); break;
    case  1: freq = 0x000020 + (offset << 1); break;
    case  2: freq = 0x000040 + (offset << 1); break;
    case  3: freq = 0x000080 + (offset << 1); break;
    case  4: freq = 0x000200 + (offset << 1); break;
    case  5: freq = 0x000400 + (offset << 2); break;
    case  6: freq = 0x001000 + (offset << 2); break;
    case  7: freq = 0x002000 + (offset << 2); break;
    case  8: freq = 0x004000 + (offset << 3); break;
    case  9: freq = 0x010000 + (offset << 3); break;
    case 10: freq = 0x020000 + (offset << 3); break;
    case 11: freq = 0x040000 + (offset << 4); break;
    case 12: freq = 0x100000 + (offset << 4); break;
    case 13: freq = 0x200000 + (offset << 4); break;
    case 14: freq = 0x400000 + (offset << 5); break;
    case 15: freq = 0x800000 + (offset << 5); break;
  }
//  DEBUG (("divider_to_freq(): segment=%d, offset=%x, freq=0x%6.6x\n",
//    segment, offset, freq));
  freq = 25000000 / freq;
//  DEBUG (("divider_to_freq(): freq=%u\n", freq));

  if (freq < 1000)
  {
    units = HZ;
    sprintf (s, "%u %s", freq, unit_name[units]);
  } else if (freq < 10000)
  {
    display = (float)freq / 1000;
    units = KHZ;
    sprintf (s, "%1.2f %s", display, unit_name[units]);
  } else if (freq < 100000)
  {
    display = (float)freq / 1000;
    units = KHZ;
    sprintf (s, "%2.1f %s", display, unit_name[units]);
  } else if (freq < 1000000)
  {
    units = KHZ;
    sprintf (s, "%d %s", freq/1000, unit_name[units]);
  } else if (freq < 10000000)
  {
    display = (float)freq / 1000000;
    units = MHZ;
    sprintf (s, "%1.2f %s", display, unit_name[units]);
  } else
  {
    display = (float)freq / 1000000;
    units = MHZ;
    sprintf (s, "%2.1f %s", display, unit_name[units]);
  }
//  DEBUG (("divider_to_freq(): string=%s\n", s));

  return freq;
}
