/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: cpu_regs.c
 *
 * Function: the CPU Register Display window
 *
 * Part of:  DIY Calculator loader application
 *
 *
 * Changelog
 *
 *  0.1  20.Oct.2006   JH   new
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>


#include <string.h>          /* strchr() */
#include <stdlib.h>          /* atoi() */

#include "diy_load.h"
#include "cpu_reg.h"
#include "serial.h"

#include "gtk_helper.h"      /* gtk_sleep() */
#include "gtk_statusbox.h"


/******************************  Defines  *********************************/


/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/**************************  Private Functions  ***************************/


/************************  Function Prototypes  ***************************/


/******************************  Globals  *********************************/
static struct CPU_REG cpu_regs;   /* the cpu register entry boxes         */
static GtkWidget *window;         /* the CPU register display window      */

extern int debug;
extern int offline;


/* cpu_reg_window_refresh */
/**************************************************************************/
void cpu_reg_window_refresh( /* refresh the CPU Register Display window   */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     :                                                            *
* called from : on clicking on the Refresh button                          *
* calls       : fill_cpu_regs()                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Refresh the CPU Register display window.                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  fill_cpu_regs ();
}

/* cpu_reg_window_destroy */
/**************************************************************************/
void cpu_reg_window_destroy (     /* destroy the CPU Reg Display window   */
  GtkWidget *widget,              /* widget which triggered the callback  */
  gpointer data)                  /* user data                            */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the CPU Register display window                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the CPU register display window.                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* do_disp_cpu_reg */
/**************************************************************************/
GtkWidget *do_disp_cpu_reg ( /* the CPU Register Display window           */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : window, cpu_regs                                           *
* called from : disp_cpu_reg()                                             *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Create the CPU register display window.                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the CPU register display window                                          *
***************************************************************************/
{
  GtkWidget *vbox, *hbox, *sr_box, *vbox1, *vbox2, *button_box;
  GtkWidget *frame, *box, *entry, *button;

  DEBUG (("do_disp_cpu_reg()\n"));
  if (window == NULL)
  {
    /*************  create the "CPU Register Display" window  *************/
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    /* install the delete_event handler that immediately exits GTK */
    g_signal_connect (G_OBJECT (window), "delete_event",
      G_CALLBACK (cpu_reg_window_destroy), NULL);
    /* set the border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (window), 10);
    /* set the window title */
    gtk_window_set_title (GTK_WINDOW (window), "CPU Register Display");
    /* when the main window will be closed, close also this one */
    gtk_window_set_destroy_with_parent (GTK_WINDOW(window), TRUE);
    /* set it on top of the main window */
    gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));

    /*******************   create the various widgets  ********************/
    vbox = gtk_vbox_new (FALSE, 4);
    /* put the box into the window */
    gtk_container_add (GTK_CONTAINER (window), vbox);

    hbox = gtk_hbox_new (TRUE, 10);
    gtk_container_add (GTK_CONTAINER (vbox), hbox);

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (hbox), vbox1);

    vbox2 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (hbox), vbox2);

    /* accumulator */
    /* the frame with title */
    frame = gtk_frame_new ("Accumulator");
    /* center the title horizontally */
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    /* put it into the left vbox */
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    /* a box within the frame (for spacing) */
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    /* the text entry */
    entry = gtk_entry_new ();
    /* read only */
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    /* default text */
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    /* center alignment */
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    /* make it a decent size */
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    /* put it into the box */
    gtk_container_add (GTK_CONTAINER (box), entry);
    /* save it in the CPU_REG object */
    cpu_regs.acc = entry;

    /* instruction register */
    frame = gtk_frame_new ("Instruction Reg");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.ir = entry;

    /* interrupt vector register */
    frame = gtk_frame_new ("Interrupt Vector");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.iv = entry;

    /* program counter */
    frame = gtk_frame_new ("Program Counter");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.pc = entry;

    /* index register */
    frame = gtk_frame_new ("Index Reg");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.x = entry;

    /* stack pointer */
    frame = gtk_frame_new ("Stack Pointer");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sp = entry;

    /* status register */
    frame = gtk_frame_new ("Status Reg");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 2);
    sr_box = gtk_hbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (sr_box), 4);
    gtk_container_add (GTK_CONTAINER (frame), sr_box);

    frame = gtk_frame_new ("I");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (sr_box), frame, TRUE, TRUE, 4);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "X");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 20, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sr_i = entry;

    frame = gtk_frame_new ("O");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (sr_box), frame, TRUE, TRUE, 4);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "X");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 20, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sr_o = entry;

    frame = gtk_frame_new ("N");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (sr_box), frame, TRUE, TRUE, 4);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "X");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 20, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sr_n = entry;

    frame = gtk_frame_new ("Z");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (sr_box), frame, TRUE, TRUE, 4);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "X");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 20, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sr_z = entry;

    frame = gtk_frame_new ("C");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (sr_box), frame, TRUE, TRUE, 4);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "X");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 20, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    cpu_regs.sr_c = entry;

    /* the buttons */
    button_box = gtk_hbox_new (TRUE, 10);
    gtk_container_add (GTK_CONTAINER (vbox), button_box);

    /* 'refresh' button */
    button = gtk_button_new_with_label ("Refresh");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(cpu_reg_window_refresh), NULL);
    gtk_box_pack_start (GTK_BOX (button_box), button, TRUE, TRUE, 0);

    /* 'close' button */
    button = gtk_button_new_with_label ("Close");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(cpu_reg_window_destroy), NULL);
    gtk_box_pack_start (GTK_BOX (button_box), button, TRUE, TRUE, 0);

    gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER_ON_PARENT);

    /* show the window */
    gtk_widget_show_all (window);
  }  /* have to create the CPU register window */

  return window;
}

/* fill_cpu_regs */
/**************************************************************************/
int fill_cpu_regs ()         /* fill the CPU register display window      */
/***************************************************************************
* globals     : cpu_regs, offline                                          *
* called from : cpu_reg_window_refresh(), disp_cpu_reg(), do_step(),       *
*               timer_callback()                                           *
* calls       : read_line(); gtk_status_box(), gtk_status_box_destroy()    *
*               gtk_status_box_set_value(), gtk_sleep(), atox()            *
* Function                                                                 *
* --------                                                                 *
* Fill the CPU register display window with values.                        *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* 1 .. timeout                                                             *
***************************************************************************/
{
  char buf[BUFLEN];
  char *p;
  unsigned char reg;
  GtkWidget *statusbox;
  int value;

  DEBUG (("fill_cpu_regs()\n"));

  statusbox = gtk_status_box (window, "I/O Ports Display",
    "Receiving I/O ports data ... ", 1);
  value = 0;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (100);

  /* program counter */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.pc), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("PC =%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.pc), p);
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* instruction register */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.ir), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("IR =%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.ir), p);
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* accumulator */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.acc), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("ACC=%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.acc), p);
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* status register */
  if (offline)
  {
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_i), "o");
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_o), "f");
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_n), "l");
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_z), "i");
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_c), "n");
  } else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      p ++;
      reg = atox (p);
      DEBUG (("SR =%2.2x\n", reg));
      /* interrupt mask bit */
      if (reg & 0x10)
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_i), "1");
      else
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_i), "0");
      /* overflow flag */
      if (reg & 0x08)
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_o), "1");
      else
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_o), "0");
      /* negative flag */
      if (reg & 0x04)
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_n), "1");
      else
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_n), "0");
      /* zero flag */
      if (reg & 0x02)
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_z), "1");
      else
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_z), "0");
      /* carry flag */
      if (reg & 0x01)
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_c), "1");
      else
        gtk_entry_set_text (GTK_ENTRY(cpu_regs.sr_c), "0");
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* index register */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.x), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("X  =%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.x), p);
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* stack pointer */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.sp), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("SP =%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.sp), p);
    }
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* interrupt vector */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(cpu_regs.iv), "offline");
  else
  {
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      DEBUG (("IV =%s\n", p));
      gtk_entry_set_text (GTK_ENTRY(cpu_regs.iv), p);
    }
    /* eat the prompt */
    if (read_line (buf, BUFLEN) == -1)
      return 1;
    if (read_line (buf, BUFLEN) == -1)
      return 1;
  }

  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);
  gtk_status_box_destroy (statusbox);

  return 0;
}
