/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: disp_mem.c
 *
 * Function: the Memory Display window
 *
 * Part of:  DIY Calculator loader application
 *
 *
 * Changelog
 *
 *  0.1  23.Oct.2006   JH   new
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include <string.h>          /* strlen() */

#include "diy_load.h"
#include "cmd.h"
#include "serial.h"

#include "gtk_helper.h"
#include "gtk_statusbox.h"


/******************************  Defines  *********************************/
enum DISPLAY_MODES {DM_HEX, DM_DISASM};
enum RELOAD_MODES {RM_AUTO, RM_YES, RM_NO};
enum CB_ACTIONS {SET_START, SET_END, SET_RANGE};

/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/
struct INSTRUCTION
{
  char name[10];
  char prefix[10];
  char suffix[10];
  int  bytes;
};

/*************************  Callback Functions  ***************************/
static void view_select (
  gpointer callback_data,
  guint callback_action,
  GtkWidget *menu_item
);

static void goto_ram (void);
static void goto_rom (void);
static void goto_input (void);
static void goto_output (void);
static void goto_stack (void);
static void goto_pc (void);
static void goto_mon (void);

void goto_address (          /* enter a new address value                 */
  GtkWidget *widget,         /* window which triggered this callback      */
  gpointer data,             /* user data                                 */
  guint action               /* callback action                           */
);

static void set_address(     /*callback function for set address dialog   */
  GtkWidget *widget,         /* the widget which triggered this callback  */
  gpointer data              /* user data                                 */
);

/**************************  Private Functions  ***************************/
static int do_display (      /* display RAM area                          */
  unsigned short start,      /* first address to display                  */
  unsigned short end,        /* last address to display                   */
  int mode,                  /* 0=hex view, 1=disasm                      */
  int reload                 /* 1=reload data                             */
);

/******************************  Globals  *********************************/
static GtkWidget *window;         /* the memory display window            */
static GtkItemFactory *mem_factory;
static GtkWidget *mem_viewer;
static GtkWidget *set_address_window;
static GtkWidget *address_entry;

extern int start_addr;
extern int end_addr;
static int dmode;            /* display mode (hex/disassembler)           */

/* the menus and entries */
static GtkItemFactoryEntry mem_items[] = {
  { "/_Goto Address",                    NULL, NULL,        0, "<Branch>"},
  { "/Goto Address/R_AM",                NULL, goto_ram,    0, "<Item>"},
  { "/Goto Address/R_OM",                NULL, goto_rom,    0, "<Item>"},
  { "/Goto Address/_Input Ports",        NULL, goto_input,  0, "<Item>"},
  { "/Goto Address/_Output Ports",       NULL, goto_output, 0, "<Item>"},
  { "/Goto Address/_Stack Pointer",      NULL, goto_stack,  0, "<Item>"},
  { "/Goto Address/_Program Counter",    NULL, goto_pc,     0, "<Item>"},
  { "/Goto Address/_Monitor RAM",        NULL, goto_mon,    0, "<Item>"},
  { "/_Enter Address",                   NULL, NULL,        0, "<Branch>"},
  { "/Enter Address/_Start Address",NULL,goto_address,SET_START,"<Item>"},
  { "/Enter Address/_End Address",  NULL,goto_address,SET_END,  "<Item>"},
  { "/Enter Address/Address _Range",NULL,goto_address,SET_RANGE,"<Item>"},
  { "/_View",                            NULL, NULL,        0, "<Branch>"},
  { "/View/He_x dump",     NULL, view_select, DM_HEX,    "<RadioItem>"},
  { "/View/_Disassembler", NULL, view_select, DM_DISASM, "/View/Hex dump"},
};

static int nmem_items = sizeof (mem_items) / sizeof (mem_items[0]);

static unsigned char data[0x10000];/* the DIY Calculator CPU memory image */

static struct INSTRUCTION inst_set[256]={
  {"NOP   ", "",   "",     1},  /* 0x00 */
  {"HALT  ", "",   "",     1},  /* 0x01 */
  {"JIR   ", "",   "",     1},  /* 0x02 */
  {"      ", "",   "",     0},  /* 0x03 */
  {"      ", "",   "",     0},  /* 0x04 */
  {"      ", "",   "",     0},  /* 0x05 */
  {"/     ", "",   "",     0},  /* 0x06 */
  {"      ", "",   "",     0},  /* 0x07 */
  {"SETIM ", "",   "",     1},  /* 0x08 */
  {"CLRIM ", "",   "",     1},  /* 0x09 */
  {"      ", "",   "",     0},  /* 0x0A */
  {"      ", "",   "",     0},  /* 0x0B */
  {"      ", "",   "",     0},  /* 0x0C */
  {"      ", "",   "",     0},  /* 0x0D */
  {"      ", "",   "",     0},  /* 0x0E */
  {"      ", "",   "",     0},  /* 0x0F */
  {"ADD   ", "$",  "",     2},  /* 0x10 */
  {"ADD   ", "[",  "]",    3},  /* 0x11 */
  {"ADD   ", "[",  ",X]",  3},  /* 0x12 */
  {"      ", "",   "",     0},  /* 0x13 */
  {"      ", "",   "",     0},  /* 0x14 */
  {"      ", "",   "",     0},  /* 0x15 */
  {"      ", "",   "",     0},  /* 0x16 */
  {"      ", "",   "",     0},  /* 0x17 */
  {"ADDC  ", "$",  "",     2},  /* 0x18 */
  {"ADDC  ", "[",  "]",    3},  /* 0x19 */
  {"ADDC  ", "[",  ",X]",  3},  /* 0x1A */
  {"      ", "",   "",     0},  /* 0x1B */
  {"      ", "",   "",     0},  /* 0x1C */
  {"      ", "",   "",     0},  /* 0x1D */
  {"      ", "",   "",     0},  /* 0x1E */
  {"      ", "",   "",     0},  /* 0x1F */
  {"SUB   ", "$",  "",     2},  /* 0x20 */
  {"SUB   ", "[",  "]",    3},  /* 0x21 */
  {"SUB   ", "[",  ",X]",  3},  /* 0x22 */
  {"      ", "",   "",     0},  /* 0x23 */
  {"      ", "",   "",     0},  /* 0x24 */
  {"      ", "",   "",     0},  /* 0x25 */
  {"      ", "",   "",     0},  /* 0x26 */
  {"      ", "",   "",     0},  /* 0x27 */
  {"SUBC  ", "$",  "",     2},  /* 0x28 */
  {"SUBC  ", "[",  "]",    3},  /* 0x29 */
  {"SUBC  ", "[",  ",X]",  3},  /* 0x2A */
  {"      ", "",   "",     0},  /* 0x2B */
  {"      ", "",   "",     0},  /* 0x2C */
  {"      ", "",   "",     0},  /* 0x2D */
  {"      ", "",   "",     0},  /* 0x2E */
  {"      ", "",   "",     0},  /* 0x2F */
  {"AND   ", "$",  "",     2},  /* 0x30 */
  {"AND   ", "[",  "]",    3},  /* 0x31 */
  {"AND   ", "[",  ",X]",  3},  /* 0x32 */
  {"      ", "",   "",     0},  /* 0x33 */
  {"      ", "",   "",     0},  /* 0x34 */
  {"      ", "",   "",     0},  /* 0x35 */
  {"      ", "",   "",     0},  /* 0x36 */
  {"      ", "",   "",     0},  /* 0x37 */
  {"OR    ", "$",  "",     2},  /* 0x38 */
  {"OR    ", "[",  "]",    3},  /* 0x39 */
  {"OR    ", "[",  ",X]",  3},  /* 0x3A */
  {"      ", "",   "",     0},  /* 0x3B */
  {"      ", "",   "",     0},  /* 0x3C */
  {"      ", "",   "",     0},  /* 0x3D */
  {"      ", "",   "",     0},  /* 0x3E */
  {"      ", "",   "",     0},  /* 0x3F */
  {"XOR   ", "$",  "",     2},  /* 0x40 */
  {"XOR   ", "[",  "]",    3},  /* 0x41 */
  {"XOR   ", "[",  ",X]",  3},  /* 0x42 */
  {"      ", "",   "",     0},  /* 0x43 */
  {"      ", "",   "",     0},  /* 0x44 */
  {"      ", "",   "",     0},  /* 0x45 */
  {"      ", "",   "",     0},  /* 0x46 */
  {"      ", "",   "",     0},  /* 0x47 */
  {"      ", "",   "",     0},  /* 0x48 */
  {"      ", "",   "",     0},  /* 0x49 */
  {"      ", "",   "",     0},  /* 0x4A */
  {"      ", "",   "",     0},  /* 0x4B */
  {"      ", "",   "",     0},  /* 0x4C */
  {"      ", "",   "",     0},  /* 0x4D */
  {"      ", "",   "",     0},  /* 0x4E */
  {"      ", "",   "",     0},  /* 0x4F */
  {"BLDSP ", "$",  "",     3},  /* 0x50 */
  {"BLDSP ", "[",  "]",    3},  /* 0x51 */
  {"      ", "",   "",     0},  /* 0x52 */
  {"      ", "",   "",     0},  /* 0x53 */
  {"      ", "",   "",     0},  /* 0x54 */
  {"      ", "",   "",     0},  /* 0x55 */
  {"      ", "",   "",     0},  /* 0x56 */
  {"      ", "",   "",     0},  /* 0x57 */
  {"      ", "",   "",     0},  /* 0x58 */
  {"BSTSP ", "[",  "]",    3},  /* 0x59 */
  {"      ", "",   "",     0},  /* 0x5A */
  {"      ", "",   "",     0},  /* 0x5B */
  {"      ", "",   "",     0},  /* 0x5C */
  {"      ", "",   "",     0},  /* 0x5D */
  {"      ", "",   "",     0},  /* 0x5E */
  {"      ", "",   "",     0},  /* 0x5F */
  {"CMPA  ", "$",  "",     2},  /* 0x60 */
  {"CMPA  ", "[",  "]",    3},  /* 0x61 */
  {"CMPA  ", "[",  ",X]",  3},  /* 0x62 */
  {"      ", "",   "",     0},  /* 0x63 */
  {"      ", "",   "",     0},  /* 0x64 */
  {"      ", "",   "",     0},  /* 0x65 */
  {"      ", "",   "",     0},  /* 0x66 */
  {"      ", "",   "",     0},  /* 0x67 */
  {"      ", "",   "",     0},  /* 0x68 */
  {"      ", "",   "",     0},  /* 0x69 */
  {"      ", "",   "",     0},  /* 0x6A */
  {"      ", "",   "",     0},  /* 0x6B */
  {"      ", "",   "",     0},  /* 0x6C */
  {"      ", "",   "",     0},  /* 0x6D */
  {"      ", "",   "",     0},  /* 0x6E */
  {"      ", "",   "",     0},  /* 0x6F */
  {"SHL   ", "",   "",     1},  /* 0x70 */
  {"SHR   ", "",   "",     1},  /* 0x71 */
  {"      ", "",   "",     0},  /* 0x72 */
  {"      ", "",   "",     0},  /* 0x73 */
  {"      ", "",   "",     0},  /* 0x74 */
  {"      ", "",   "",     0},  /* 0x75 */
  {"      ", "",   "",     0},  /* 0x76 */
  {"      ", "",   "",     0},  /* 0x77 */
  {"ROLC  ", "",   "",     1},  /* 0x78 */
  {"RORC  ", "",   "",     1},  /* 0x79 */
  {"      ", "",   "",     0},  /* 0x7A */
  {"      ", "",   "",     0},  /* 0x7B */
  {"      ", "",   "",     0},  /* 0x7C */
  {"      ", "",   "",     0},  /* 0x7D */
  {"      ", "",   "",     0},  /* 0x7E */
  {"      ", "",   "",     0},  /* 0x7F */
  {"INCA  ", "",   "",     1},  /* 0x80 */
  {"DECA  ", "",   "",     1},  /* 0x81 */
  {"INCX  ", "",   "",     1},  /* 0x82 */
  {"DECX  ", "",   "",     1},  /* 0x83 */
  {"      ", "",   "",     0},  /* 0x84 */
  {"      ", "",   "",     0},  /* 0x85 */
  {"      ", "",   "",     0},  /* 0x86 */
  {"      ", "",   "",     0},  /* 0x87 */
  {"      ", "",   "",     0},  /* 0x88 */
  {"      ", "",   "",     0},  /* 0x89 */
  {"      ", "",   "",     0},  /* 0x8A */
  {"      ", "",   "",     0},  /* 0x8B */
  {"      ", "",   "",     0},  /* 0x8C */
  {"      ", "",   "",     0},  /* 0x8D */
  {"      ", "",   "",     0},  /* 0x8E */
  {"      ", "",   "",     0},  /* 0x8F */
  {"LDA   ", "$",  "",     2},  /* 0x90 */
  {"LDA   ", "[",  "]",    3},  /* 0x91 */
  {"LDA   ", "[",  ",X]",  3},  /* 0x92 */
  {"LDA   ", "[[", "]]",   3},  /* 0x93 */
  {"LDA   ", "[[", ",X]]", 3},  /* 0x94 */
  {"LDA   ", "[[", "],X]", 3},  /* 0x95 */
  {"      ", "",   "",     0},  /* 0x96 */
  {"      ", "",   "",     0},  /* 0x97 */
  {"VER   ", "",   "",     1},  /* 0x98 */
  {"STA   ", "[",  "]",    3},  /* 0x99 */
  {"STA   ", "[",  ",X]",  3},  /* 0x9A */
  {"STA   ", "[[", "]]",   3},  /* 0x9B */
  {"STA   ", "[[", ",X]]", 3},  /* 0x9C */
  {"STA   ", "[[", "],X]", 3},  /* 0x9D */
  {"      ", "",   "",     0},  /* 0x9E */
  {"      ", "",   "",     0},  /* 0x9F */
  {"BLDX  ", "$",  "",     3},  /* 0xA0 */
  {"BLDX  ", "[",  "]",    3},  /* 0xA1 */
  {"      ", "",   "",     0},  /* 0xA2 */
  {"      ", "",   "",     0},  /* 0xA3 */
  {"      ", "",   "",     0},  /* 0xA4 */
  {"      ", "",   "",     0},  /* 0xA5 */
  {"      ", "",   "",     0},  /* 0xA6 */
  {"      ", "",   "",     0},  /* 0xA7 */
  {"      ", "",   "",     0},  /* 0xA8 */
  {"BSTX  ", "[",  "]",    3},  /* 0xA9 */
  {"      ", "",   "",     0},  /* 0xAA */
  {"      ", "",   "",     0},  /* 0xAB */
  {"      ", "",   "",     0},  /* 0xAC */
  {"      ", "",   "",     0},  /* 0xAD */
  {"      ", "",   "",     0},  /* 0xAE */
  {"      ", "",   "",     0},  /* 0xAF */
  {"POPA  ", "",   "",     1},  /* 0xB0 */
  {"POPSR ", "",   "",     1},  /* 0xB1 */
  {"PUSHA ", "",   "",     1},  /* 0xB2 */
  {"PUSHSR", "",   "",     1},  /* 0xB3 */
  {"      ", "",   "",     0},  /* 0xB4 */
  {"      ", "",   "",     0},  /* 0xB5 */
  {"      ", "",   "",     0},  /* 0xB6 */
  {"      ", "",   "",     0},  /* 0xB7 */
  {"      ", "",   "",     0},  /* 0xB8 */
  {"      ", "",   "",     0},  /* 0xB9 */
  {"      ", "",   "",     0},  /* 0xBA */
  {"      ", "",   "",     0},  /* 0xBB */
  {"      ", "",   "",     0},  /* 0xBC */
  {"      ", "",   "",     0},  /* 0xBD */
  {"      ", "",   "",     0},  /* 0xBE */
  {"      ", "",   "",     0},  /* 0xBF */
  {"      ", "",   "",     0},  /* 0xC0 */
  {"JMP   ", "[",  "]",    3},  /* 0xC1 */
  {"JMP   ", "[",  ",X]",  3},  /* 0xC2 */
  {"JMP   ", "[[", "]]",   3},  /* 0xC3 */
  {"JMP   ", "[[", ",X]]", 3},  /* 0xC4 */
  {"JMP   ", "[[", "],X]", 3},  /* 0xC5 */
  {"      ", "",   "",     0},  /* 0xC6 */
  {"RTI   ", "",   "",     1},  /* 0xC7 */
  {"      ", "",   "",     0},  /* 0xC8 */
  {"JSR   ", "[",  "]",    3},  /* 0xC9 */
  {"JSR   ", "[",  ",X]",  3},  /* 0xCA */
  {"JSR   ", "[[", "]]",   3},  /* 0xCB */
  {"JSR   ", "[[", ",X]]", 3},  /* 0xCC */
  {"JSR   ", "[[", "],X]", 3},  /* 0xCD */
  {"      ", "",   "",     0},  /* 0xCE */
  {"RTS   ", "",   "",     1},  /* 0xCF */
  {"      ", "",   "",     0},  /* 0xD0 */
  {"JZ    ", "[",  "]",    3},  /* 0xD1 */
  {"      ", "",   "",     0},  /* 0xD2 */
  {"      ", "",   "",     0},  /* 0xD3 */
  {"      ", "",   "",     0},  /* 0xD4 */
  {"      ", "",   "",     0},  /* 0xD5 */
  {"JNZ   ", "[",  "]",    3},  /* 0xD6 */
  {"      ", "",   "",     0},  /* 0xD7 */
  {"      ", "",   "",     0},  /* 0xD8 */
  {"JN    ", "[",  "]",    3},  /* 0xD9 */
  {"      ", "",   "",     0},  /* 0xDA */
  {"      ", "",   "",     0},  /* 0xDB */
  {"      ", "",   "",     0},  /* 0xDC */
  {"      ", "",   "",     0},  /* 0xDD */
  {"JNN   ", "[",  "]",    3},  /* 0xDE */
  {"      ", "",   "",     0},  /* 0xDF */
  {"      ", "",   "",     0},  /* 0xE0 */
  {"JC    ", "[",  "]",    3},  /* 0xE1 */
  {"      ", "",   "",     0},  /* 0xE2 */
  {"      ", "",   "",     0},  /* 0xE3 */
  {"      ", "",   "",     0},  /* 0xE4 */
  {"      ", "",   "",     0},  /* 0xE5 */
  {"JNC   ", "[",  "]",    3},  /* 0xE6 */
  {"      ", "",   "",     0},  /* 0xE7 */
  {"      ", "",   "",     0},  /* 0xE8 */
  {"JO    ", "[",  "]",    3},  /* 0xE9 */
  {"      ", "",   "",     0},  /* 0xEA */
  {"      ", "",   "",     0},  /* 0xEB */
  {"      ", "",   "",     0},  /* 0xEC */
  {"      ", "",   "",     0},  /* 0xED */
  {"JNO   ", "[",  "]",    3},  /* 0xEE */
  {"      ", "",   "",     0},  /* 0xEF */
  {"BLDIV ", "$",  "",     3},  /* 0xF0 */
  {"BLDIV ", "[",  "]",    3},  /* 0xF1 */
  {"      ", "",   "",     0},  /* 0xF2 */
  {"      ", "",   "",     0},  /* 0xF3 */
  {"      ", "",   "",     0},  /* 0xF4 */
  {"      ", "",   "",     0},  /* 0xF5 */
  {"      ", "",   "",     0},  /* 0xF6 */
  {"      ", "",   "",     0},  /* 0xF7 */
  {"      ", "",   "",     0},  /* 0xF8 */
  {"BSTIV ", "[",  "]",    3},  /* 0xF9 */
  {"      ", "",   "",     0},  /* 0xFA */
  {"      ", "",   "",     0},  /* 0xFB */
  {"      ", "",   "",     0},  /* 0xFC */
  {"      ", "",   "",     0},  /* 0xFD */
  {"      ", "",   "",     0},  /* 0xFE */
  {"      ", "",   "",     0},  /* 0xFF */
};

extern int debug;
extern int offline;
extern int do_timer_cb;


/* mem_window_destroy */
/**************************************************************************/
void mem_window_destroy (    /* destroy the memory display window         */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the memory display window                       *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the memory display window.                                       *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* do_disp_mem */
/**************************************************************************/
GtkWidget *do_disp_mem (     /* the memory display window                 */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : window, dmode, start_addr, end_addr, mem_factory, data[]   *
*               mem_viewer                                                 *
* called from : disp_mem()                                                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the contents of the CPU registers.                               *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* 1 .. timeout                                                             *
***************************************************************************/
{
  GtkWidget *overall_box;
  GtkWidget *button;
  GtkWidget *button_box;
  GtkWidget *menubar;
  GtkAccelGroup *mem_accel;
  GtkWidget *text_box;
  GtkTextBuffer *buffer;
  PangoFontDescription *font;
//  GdkColor color;

  DEBUG (("do_disp_mem()\n"));

  if (window == NULL)
  {
    /* clear the DIY-Calculator CPU memory image */
    memset (data, 0, sizeof(data));

    /***************  create the "Memory Display" window  *****************/
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    /* install the delete_event handler that immediately exits GTK */
    g_signal_connect (G_OBJECT (window), "delete_event",
      G_CALLBACK (mem_window_destroy), NULL);
    /* set the border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (window), 10);
    /* set the window title */
    gtk_window_set_title (GTK_WINDOW (window), "Memory Display");
    /* when the main window will be closed, close also this one */
    gtk_window_set_destroy_with_parent (GTK_WINDOW(window), TRUE);
    /* set it on top of the main window */
    gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));

    /* first create an overall box to arrange all other widgets */
    overall_box = gtk_vbox_new (FALSE, 0);

    /* put the box into the main window */
    gtk_container_add (GTK_CONTAINER (window), overall_box);

    /******************************  the menu  ****************************/
    /* Make an accelerator group (shortcut keys) */
    mem_accel = gtk_accel_group_new ();

    /* Make an ItemFactory (that makes a menubar) */
    mem_factory=gtk_item_factory_new(GTK_TYPE_MENU_BAR,"<main>",mem_accel);
    /* This function generates the menu items. Pass the item factory,
       the number of items in the array, the array itself, and any
       callback data for the the menu items. */
    gtk_item_factory_create_items (mem_factory, nmem_items, mem_items, NULL);

    /* Attach the new accelerator group to the window. */
    gtk_window_add_accel_group (GTK_WINDOW (window), mem_accel);

    /* Finally, return the actual menu bar created by the item factory. */
    menubar = gtk_item_factory_get_widget (mem_factory, "<main>");
    gtk_box_pack_start (GTK_BOX (overall_box), menubar, FALSE, FALSE, 2);

    /*************************  create the text box  **********************/
    /* create scrollable window for the text box */
    text_box = gtk_scrolled_window_new (FALSE, FALSE);
    gtk_container_add (GTK_CONTAINER(overall_box), text_box);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (text_box),
                                    GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);
    /* set an inside border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (text_box), 5);
    /* create the TextView widget */
    mem_viewer = gtk_text_view_new ();
    gtk_container_add (GTK_CONTAINER(text_box), mem_viewer);
    /* obtain the text buffer object from the text view object */
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (mem_viewer));
    /* set monospace font */
    font = pango_font_description_from_string (TERMINAL_FONT);
    gtk_widget_modify_font (mem_viewer, font);
  
    /* set black background */
//  gdk_color_parse ("black", &color);
//  gtk_widget_modify_base (mem_viewer, GTK_STATE_NORMAL, &color);
    gtk_widget_set_size_request (window, 640, 450);

    /**************************  the buttons  *****************************/
    /* create a box for the buttons */
    button_box = gtk_hbox_new (TRUE, 20);

    /* put into the window */
    gtk_box_pack_start (GTK_BOX (overall_box), button_box, FALSE, TRUE, 0);

    /* 'Start Address' button */
    button = gtk_button_new_with_label ("Start Address");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(goto_address), (gpointer)SET_START);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    /* 'End Address' button */
    button = gtk_button_new_with_label ("End Address");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(goto_address), (gpointer)SET_END);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    /* 'Address Range' button */
    button = gtk_button_new_with_label ("Address Range");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(goto_address), (gpointer)SET_RANGE);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    /* 'Close' button */
    button = gtk_button_new_with_label ("Close");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(mem_window_destroy), NULL);
    gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

    /* show everything */
    gtk_widget_show_all (window);

    dmode = DM_HEX;
  }
  do_display (start_addr, end_addr, dmode, RM_YES);

  return window;
}

/* do_display */
/**************************************************************************/
int do_display (             /* display RAM area                          */
  unsigned short start,      /* first address to display                  */
  unsigned short end,        /* last address to display                   */
  int mode,                  /* 0=hex view, 1=disasm                      */
  int reload)                /* 1=reload data                             */
/***************************************************************************
* globals     : window, offline, do_timer_cb, mem_viewer,data[],inst_set[] *
* called from : do_disp_mem(), view_select(), goto_ram(), goto_rom(),      *
*               goto_stack(), goto_pc(), set_address()                     *
* calls       : write_port(), read_line(), gtk_status_box(), gtk_is_print()*
*               gtk_status_box_set_value(), gtk_status_box_destroy()       *
*               gtk_text_view_printf()                                     *
* Function                                                                 *
* --------                                                                 *
* Display an address area.                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
***************************************************************************/
{
  char buf[BUFLEN];
  unsigned int addr;
  char *p;
  int i, bytes_in_line;
  unsigned char cur_inst;
  static unsigned short old_start, old_end;
  GtkWidget *statusbox;
  int value, lineno, delta, no_of_lines;

  DEBUG (("do_disp_mem(): start=0x%4.4x, end=0x%4.4x, mode=%d, reload=%d\n",
    start, end, mode, reload));
  /* decode the reload mode */
  switch (reload)
  {
    case RM_AUTO:
//DEBUG (("do_disp_mem(): auto reload, old_start=0x%4.4X, old_end=0x%4.4X\n",
//        old_start, old_end));
//DEBUG (("                                start=0x%4.4X,     end=0x%4.4X\n",
//        start, end));
      if (old_start <= start && old_end >= end)
        reload = 0;
      else
      {
        reload = 1;
        old_start = start;
        old_end = end;
      }
      break;
    case RM_YES:
      reload = 1;
      old_start = start;
      old_end = end;
      break;
    case RM_NO:
      reload = 0;
      break;
  }
  if (reload && !offline)
  {
    /* send the "memory display" command */
    sprintf (buf, "md %x %x\n", start_addr, end_addr-start_addr+1);
    DEBUG (("do_disp_mem(): sending cmd \"%s\"\n", buf));
    do_timer_cb = 0;
    write_port (buf, strlen(buf));
    usleep (200000);
    /* command line */
    read_line (buf, BUFLEN);
    /* empty line */
    read_line (buf, BUFLEN);
    /* empty line */
    read_line (buf, BUFLEN);
    /* title line */
    read_line (buf, BUFLEN);
    /* receive all lines and put the contents into a buffer */
    addr = start_addr;
    statusbox=gtk_status_box(window,"Memory Display","Receiving data ...",1);
    lineno = 0;
    value = 0;
    no_of_lines = (end - start) / 16;
    if (no_of_lines == 0)
      no_of_lines ++;
    if (no_of_lines > 100)
      delta = no_of_lines / 100;
    else
      delta = 100 / no_of_lines;
    gtk_status_box_set_value (statusbox, value);
    DEBUG (("do_display(): no_of_lines=%d, delta=%d\n",no_of_lines,delta));
    do
    {
      if (read_line (buf, BUFLEN) <= 0)
        break;
      /* skip the address at the begin of the line */
      p = buf + 6;
      /* compute the number of bytes in this line from the start address */
      bytes_in_line = 16 - (addr & 0xf);
//      DEBUG (("do_display(): addr=0x%4.4X, bytes_in_line=%d, line=%s",
//        addr,bytes_in_line, buf));
      for (i=0; i<bytes_in_line && addr+i<=end; i++)
      {
        data[addr+i] = atox (p);
        p += 3;
        if (i == bytes_in_line-9)
          p ++;
      }
      addr += i;
      lineno ++;
      if ((no_of_lines>100 && (lineno%delta)==0) || no_of_lines<=100)
      {
        if (no_of_lines <= 100)
          value += delta;
        else
          value ++;
        if (value > 100)
          value = 100;
//        DEBUG(("do_display(): updating progressbar: lineno=%d, value=%d\n",
//          lineno, value));
        gtk_status_box_set_value (statusbox, value);
      }
      gtk_sleep (1);
    } while (addr <= end);
    gtk_status_box_destroy (statusbox);
    do_timer_cb = 1;
  }
  gtk_text_view_printf (mem_viewer, "\n");
  switch (mode)
  {
    case DM_HEX:      /* hex dump view */
      for (addr=start; addr<end; addr+=16)
      {
        /* print the address */
        gtk_text_view_printf (mem_viewer, "%4.4X ", addr);
        /* print the hex bytes */
        for (i=0; i<16 && addr+i<=end; i++)
        {
          gtk_text_view_printf (mem_viewer, " %2.2X", data[addr+i]);
          if (i == 7)
            gtk_text_view_printf (mem_viewer, " ");
        }
        /* fill the line with spaces */
        for ( ; i<16; i++)
          gtk_text_view_printf (mem_viewer, "   ");
        /* space to the ASCII part */
        gtk_text_view_printf (mem_viewer, "  ");
        /* print ASCII */
        for (i=0; i<16 && addr+i<=end; i++)
        {
          if (gtk_isprint (data[addr+i]))
            gtk_text_view_printf (mem_viewer, "%c", data[addr+i]);
          else
            gtk_text_view_printf (mem_viewer, ".");
          if (i == 7)
            gtk_text_view_printf (mem_viewer, " ");
        }
        gtk_text_view_printf (mem_viewer, "\n");
      }
      break;

    case DM_DISASM:   /* disassembler view */
      addr = start;
      do 
      {
        cur_inst = data[addr];
//        DEBUG (("addr==x%4.4x, curr_inst=0x%2.2x, bytes=%d\n",
//          addr, cur_inst, inst_set[cur_inst].bytes));
        gtk_text_view_printf (mem_viewer, "%4.4X  ", addr);
        addr ++;
        if (inst_set[cur_inst].bytes > 0)
        {
          gtk_text_view_printf (mem_viewer, "%s %s",
            inst_set[cur_inst].name, inst_set[cur_inst].prefix);
          for (i=1; i<inst_set[cur_inst].bytes; i++)
          {
            gtk_text_view_printf (mem_viewer, "%2.2X", data[addr]);
            addr ++;
          }
          gtk_text_view_printf(mem_viewer,"%s\n",inst_set[cur_inst].suffix);
        } else
          gtk_text_view_printf (mem_viewer, ".BYTE %2.2X\n", cur_inst);
      } while (addr <= end); 
      break;
  }
//  DEBUG (("do_display(): leave\n"));
  return 0;
}

/*************************  Callback Functions  ***************************/
/* view_select */
/**************************************************************************/
void view_select (           /* select the display mode                   */
  gpointer callback_data,    /*                                           */
  guint callback_action,     /* the value associated with this menu item  */
  GtkWidget *menu_item )     /* which menu item triggered this callback   */
/***************************************************************************
* globals     : start_addr, end_addr, dmode                                *
* called from : menu "/View/Hex dump", "/View/Disassembler"                *
* calls       : do_display()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Select the display mode.                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  if (GTK_CHECK_MENU_ITEM(menu_item)->active)
  {
    dmode = callback_action;
    DEBUG (("view_select(): dmode=%d\n", dmode));
    do_display (start_addr, end_addr, dmode, RM_AUTO);
  }
}

/* goto_ram */
/**************************************************************************/
void goto_ram ()             /* display RAM area                          */
/***************************************************************************
* globals     : start_addr, end_addr, dmode                                *
* called from : menu "/Goto Address/RAM"                                   *
* calls       : do_dsiplay()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the RAM area.                                                    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  int range;

  range = end_addr - start_addr;
  start_addr = 0x4000;
  end_addr = start_addr + range;
  do_display (start_addr, end_addr, dmode, RM_AUTO);
}

/* goto_rom */
/**************************************************************************/
void goto_rom ()             /* display ROM area                          */
/***************************************************************************
* globals     : start_addr, end_addr, dmode                                *
* called from : menu "/Goto Address/ROM"                                   *
* calls       : do_display()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the ROM area.                                                    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  int range;

  range = end_addr - start_addr;
  start_addr = 0x0000;
  end_addr = start_addr + range;
  do_display (start_addr, end_addr, dmode, RM_AUTO);
}

/* goto_input */
/**************************************************************************/
void goto_input ()           /* display input port address range          */
/***************************************************************************
* globals     : start_addr, end_addr, dmode                                *
* called from : menu "/Goto Address/Input Ports"                           *
* calls       : do_display()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the input ports address range.                                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  GtkWidget *menu_item;

  start_addr = 0xf000;
  end_addr = 0xf01f;
  dmode = DM_HEX;
  menu_item = gtk_item_factory_get_item (
                GTK_ITEM_FACTORY(mem_factory), "/View/Hex dump");
  gtk_menu_item_activate (GTK_MENU_ITEM(menu_item));
}

/* goto_output */
/**************************************************************************/
void goto_output ()          /* display output port address range         */
/***************************************************************************
* globals     : start_addr, end_addr, dmode, mem_factory                   *
* called from : menu "/Goto Address/Output Ports"                          *
* calls       : view_select() via menu "/View/Hex dump"                    *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the output port address range.                                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  GtkWidget *menu_item;

  start_addr = 0xf020;
  end_addr = 0xf03f;
  dmode = DM_HEX;
  menu_item = gtk_item_factory_get_item (
                GTK_ITEM_FACTORY(mem_factory), "/View/Hex dump");
  gtk_menu_item_activate (GTK_MENU_ITEM(menu_item));
}

/* goto_stack */
/**************************************************************************/
void goto_stack ()           /* display stack address area                */
/***************************************************************************
* globals     : start_addr, end_addr, dmode, do_timer_cb                   *
* called from : menu "/Goto Address/Stack Pointer"                         *
* calls       : send_cmd(), read_line(), atox(), do_display()              *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the stack area.                                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char buf[BUFLEN];
  char *p;
  int range;

  if (offline)
  {
    start_addr = 0xeff0;
    end_addr = 0xf000;
    do_display (start_addr, end_addr, dmode, RM_YES);
  } else
  {
    do_timer_cb = 0;
    send_cmd (DUMPREG);
    usleep(300000);
  
    /* wait for title line */
    do
      read_line (buf, BUFLEN);
    while (strstr(buf, "register dump") == NULL);
  
    /* program counter */
    read_line (buf, BUFLEN);
    /* instruction register */
    read_line (buf, BUFLEN);
    /* accumulator */
    read_line (buf, BUFLEN);
    /* status register */
    read_line (buf, BUFLEN);
    /* index register */
    read_line (buf, BUFLEN);
    /* stack pointer */
    read_line (buf, BUFLEN);
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      range = end_addr - start_addr;
      p ++;
      start_addr = atox (p);
      DEBUG (("goto_stack(): SP=0x%4.4x\n", start_addr));
      end_addr = start_addr + range;
      if (end_addr > 0xf000)
        end_addr = 0xf000;
  
      /* eat remaining lines of rdump cmd */
      /* interrupt vector */
      read_line (buf, BUFLEN);
      /* empty line */
      read_line (buf, BUFLEN);
  
      do_display (start_addr, end_addr, dmode, RM_YES);
    }
    do_timer_cb = 1;
  }
}

/* goto_pc */
/**************************************************************************/
void goto_pc ()              /* display current program counter area      */
/***************************************************************************
* globals     : start_addr, end_addr, dmode, do_timer_cb, mem_factory      *
* called from : menu "/Goto Address/Program Counter"                       *
* calls       : do_display(), send_cmd(), read_line(), atox(),             *
*               view_select via menu "/View/Disassembler"                  *
* Function                                                                 *
* --------                                                                 *
* Display the memory beginningat the current program couinter              *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char buf[BUFLEN];
  char *p;
  int range;
  GtkWidget *menu_item;

  if (offline)
  {
    start_addr = 0x4000;
    end_addr = 0x40ff;
    do_display (start_addr, end_addr, dmode, RM_YES);
  } else
  {
    do_timer_cb = 0;
    send_cmd (DUMPREG);
    usleep(300000);
  
    /* wait for title line */
    do
      read_line (buf, BUFLEN);
    while (strstr(buf, "register dump") == NULL);
  
    /* program counter */
    read_line (buf, BUFLEN);
    p = strchr (buf, '\r');
    if (p)
      *p = '\0';
    p = strchr (buf, '$');
    if (p)
    {
      range = end_addr - start_addr;
      p ++;
      start_addr = atox (p);
      DEBUG (("goto_pc(): PC=0x%4.4x\n", start_addr));
      end_addr = start_addr + range;
  
      /* eat remaining lines of rdump cmd */
      /* instruction register */
      read_line (buf, BUFLEN);
      /* accumulator */
      read_line (buf, BUFLEN);
      /* status register */
      read_line (buf, BUFLEN);
      /* index register */
      read_line (buf, BUFLEN);
      /* stack pointer */
      read_line (buf, BUFLEN);
      /* interrupt vector */
      read_line (buf, BUFLEN);
      /* empty line */
      read_line (buf, BUFLEN);
    
      dmode = DM_DISASM;
      menu_item = gtk_item_factory_get_item (
                    GTK_ITEM_FACTORY(mem_factory), "/View/Disassembler");
      gtk_menu_item_activate (GTK_MENU_ITEM(menu_item));
    }
    do_timer_cb = 1;
  }
}

/* goto_mon */
/**************************************************************************/
void goto_mon ()             /* display ROM monitor address range         */
/***************************************************************************
* globals     : start_addr, end_addr, dmode, mem_factory                   *
* called from : menu "/Goto Address/Monitor RAM"                           *
* calls       : view_select() via menu "/View/Hex dump"                    *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the ROM monitor address range.                                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  GtkWidget *menu_item;

  start_addr = 0xfe00;
  end_addr = 0xfffe;    /* should be 0xffff, bug in ROM monitor */
  dmode = DM_HEX;
  menu_item = gtk_item_factory_get_item (
                GTK_ITEM_FACTORY(mem_factory), "/View/Hex dump");
  gtk_menu_item_activate (GTK_MENU_ITEM(menu_item));
}

/* goto_address */
/**************************************************************************/
void goto_address (          /* enter a new address value                 */
  GtkWidget *widget,         /* window which triggered this callback      */
  gpointer data,             /* user data                                 */
  guint action)              /* callback action                           */
/***************************************************************************
* globals     : start_addr, end_addr, set_address_window, address_entry    *
* called from : menu "/Enter Address/Start Address", "/Enter Address/End   *
*               Address", "/Enter Address/Address Range"                   *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Enter a new value for start address/end address/address range.           *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  GtkWidget *table, *button, *label;
  static const char *names[]={"Start Address","End Address","Address Range"};
  char buf[BUFLEN];
  int value;

  DEBUG(("goto_address(); action=%d, data=%d\n",action,(unsigned int)data));
  action = (unsigned int) data;
  strcpy (buf, "Set ");
  strcat (buf, names[action]);
  set_address_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_modal (GTK_WINDOW(set_address_window), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(set_address_window),
    GTK_WINDOW(window));
  gtk_window_set_title (GTK_WINDOW (set_address_window), buf);
  gtk_container_set_border_width (GTK_CONTAINER (set_address_window), 10);

  table = gtk_table_new (2, 2, TRUE);
  gtk_table_set_row_spacings (GTK_TABLE(table), 10);
  gtk_table_set_col_spacings (GTK_TABLE(table), 10);
  gtk_container_add (GTK_CONTAINER(set_address_window), table);
  
  strcpy (buf, names[action]);
  strcat (buf, " $");
  label = gtk_label_new (buf);
  gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);

  value = 0;
  switch (action)
  {
    case SET_START: value = start_addr; break;
    case SET_END:   value = end_addr;   break;
    case SET_RANGE: value = end_addr - start_addr; value ++; break;
  }
  sprintf (buf, "%X", value);
  action <<= 4;
  address_entry = gtk_entry_new ();
  g_signal_connect (G_OBJECT (address_entry), "activate",
        G_CALLBACK (set_address), (gpointer) (1|action));
  gtk_widget_set_size_request (GTK_WIDGET(address_entry), 60, 20);
  gtk_entry_set_text (GTK_ENTRY(address_entry), buf);
  gtk_table_attach_defaults (GTK_TABLE(table), address_entry, 1, 2, 0, 1);

  /* Set and Cancel buttons */
  button = gtk_button_new_with_label ("Set");
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (set_address), (gpointer)(1|action));
  gtk_table_attach_defaults (GTK_TABLE (table), button, 0, 1, 1, 2);

  button = gtk_button_new_with_label ("Cancel");
  g_signal_connect (G_OBJECT (button), "clicked",
		    G_CALLBACK (set_address), (gpointer)(2|action));
  gtk_table_attach_defaults (GTK_TABLE (table), button, 1, 2, 1, 2);

  /* put the dialog box to the center of the main window */
  gtk_window_set_position (GTK_WINDOW (set_address_window),
    GTK_WIN_POS_CENTER_ON_PARENT);

  gtk_widget_show_all (set_address_window);
  DEBUG (("goto_address(): leaving\n"));
}

/* set_address */
/**************************************************************************/
void set_address (           /* callback function for set address dialog  */
  GtkWidget *widget,         /* the widget which triggered this callback  */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : start_addr,end_addr,dmode,set_address_window,address_entry *
* called from : Set button / Cancel button in "Set Address" dialog         *
* calls       : do_display()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Callback function for the buttons inside the "Find Address" dialog.      *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  int action, button, value, range;
  char *p;

  DEBUG (("set_address(): data=0x%2.2x\n", (int)data));
  action = (int)data >> 4;
  button = (int)data & 0xf;
  if (button == 1)
  {
    p = (char *)gtk_entry_get_text (GTK_ENTRY(address_entry));
    value = atox (p);
    switch (action)
    {
      case SET_START:
        range = end_addr - start_addr;
        start_addr = value;
        end_addr = start_addr + range;
        break;
      case SET_END:
        end_addr = value;
        break;
      case SET_RANGE:
        end_addr = start_addr + value;
        end_addr --;
        break;
    }
    DEBUG (("set_address(): start_addr=0x%4.4x, end_addr=0x%4.4x\n",
      start_addr, end_addr));
  }
  gtk_widget_destroy (set_address_window);
  if (button == 1)
    do_display (start_addr, end_addr, dmode, RM_AUTO);
}
