/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: gtk_helper.c
 *
 * Function: some helper functions for GTK
 *
 * Part of:  DIY Calculator loader application
 *
 * Changelog
 *
 *  0.1  13.Oct.2006   JH   new
 *
 */

#include <gtk/gtk.h>

#include <string.h>
#include <stdio.h>

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>          /* sleep(), usleep() */
#endif


/******************************  Defines  *********************************/
#define BUFLEN      200

/******************************  Typedefs *********************************/

/********************************  Macros  ********************************/

/**************************  Private Functions  ***************************/

/******************************  Globals  *********************************/


/* gtk_text_view_printf */
/**************************************************************************/
int gtk_text_view_printf (   /* printf() to a GTK TextView                */
  GtkWidget *text_view,      /* the TextView                              */
  const char *format,        /* format string                             */
  ...)                       /* variable argument list                    */
/***************************************************************************
* globals     :                                                            *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* printf() to a GTK TextView.                                              *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of characters printed                                             *
***************************************************************************/
{
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  GtkTextMark *last_pos;
  va_list ap;
  char buf[BUFLEN];

  if (text_view)
  {
    va_start (ap, format);
    vsnprintf (buf, BUFLEN-1, format, ap);
    va_end (ap);
  
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));
    /* output the formatted string */
    gtk_text_buffer_insert_at_cursor (buffer, buf, -1);
    /* scroll to the end */
    gtk_text_buffer_get_end_iter (buffer, &iter);
    last_pos = gtk_text_buffer_create_mark (
      buffer, "last_pos", &iter, FALSE);
    gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW(text_view), last_pos);

    return strlen (buf);
  } else
    return 0;
}

/* gtk_text_view_puts */
/**************************************************************************/
int gtk_text_view_puts (     /* puts() to a GTK TextView                  */
  GtkWidget *text_view,      /* the TextView                              */
  const char *s)             /* the string                                */
/***************************************************************************
* globals     :                                                            *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* puts() to a GTK TextView.                                                *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of characters printed                                             *
***************************************************************************/
{
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  GtkTextMark *last_pos;
  char buf[BUFLEN];

  if (text_view)
  {
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));
    /* output the string */
    gtk_text_buffer_insert_at_cursor (buffer, buf, -1);
    /* scroll to the end */
    gtk_text_buffer_get_end_iter (buffer, &iter);
    last_pos = gtk_text_buffer_create_mark (
      buffer, "last_pos", &iter, FALSE);
    gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW(text_view), last_pos);

    return strlen (buf);
  } else
    return 0;
}

/* gtk_text_view_putchar */
/**************************************************************************/
int gtk_text_view_putchar (  /* putchar() to a GTK TextView               */
  GtkWidget *text_view,      /* the TextView                              */
  int c)                     /* the character we want to print            */
/***************************************************************************
* globals     :                                                            *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* putchar() to a GTK TextView.                                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the character                                                            *
***************************************************************************/
{
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  GtkTextMark *last_pos;
  char buf[2];

  if (text_view)
  {
    buf[0] = (char)c;
    buf[1] = '\0';
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));
    /* output the character */
    gtk_text_buffer_insert_at_cursor (buffer, buf, -1);
    /* scroll to the end */
    gtk_text_buffer_get_end_iter (buffer, &iter);
    last_pos = gtk_text_buffer_create_mark (
      buffer, "last_pos", &iter, FALSE);
    gtk_text_view_scroll_mark_onscreen (GTK_TEXT_VIEW(text_view), last_pos);
  }
  return c;
}

/* gtk_text_view_getch */
/**************************************************************************/
int gtk_text_view_getch (    /* getch() from a GTK TextView               */
  GtkWidget *text_view)      /* the TextView                              */
/***************************************************************************
* globals     :                                                            *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* getch() from a GTK TextView.                                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the character                                                            *
***************************************************************************/
{
  GtkTextBuffer *buffer;
  GtkTextIter iter1, iter2;
  static int char_offset, old_char_offset;
  char c;

  c = '\0';
  if (text_view)
  {
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text_view));
    /* set the iters to the end of the buffer (current cursor position?) */
    gtk_text_buffer_get_end_iter (buffer, &iter1);
    /* check for backspace */
    char_offset = gtk_text_iter_get_offset (&iter1);
    if (char_offset < old_char_offset)
      c = '\b';
    else
    {
      /* go back one char position */
      gtk_text_iter_backward_chars (&iter1, 1);
      /* get the character */
      c = gtk_text_iter_get_char (&iter1);
      /* delete the character from the text buffer (no echo) */
      memcpy (&iter2, &iter1, sizeof(iter1));
      gtk_text_iter_forward_chars (&iter2, 1);
      gtk_text_buffer_delete (buffer, &iter1, &iter2);
    }
  
    old_char_offset = char_offset;
  }
  return c;
}

/* gtk_set_cursor */
/**************************************************************************/
void gtk_set_cursor (        /* set a new cursor shape                    */
  GtkWidget *window,         /* in which window                           */
  GdkCursorType type)        /* which type (uses GDK_.. constants)        */
/***************************************************************************
* globals     :                                                            *
* called from : main_callback()                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Set a new cursor type (uses predefined GDK_... constants from            *
* /usr/include/gtk-1.2/gdk/gdkcursors.h                                    *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
  GdkCursor *cursor;

  cursor = gdk_cursor_new (type);
  gdk_window_set_cursor (window->window, cursor);
  gdk_cursor_destroy (cursor);
  gdk_flush ();
}

/* gtk_sleep */
/**************************************************************************/
void gtk_sleep (             /* sleep a number of msec in GTK             */
  unsigned int ms)           /* how long                                  */
/***************************************************************************
* globals     :                                                            *
* called from : main_callback(), config_callback()                         *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Disconnect from the scheduler and do some GTK update stuff.              *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
  int i;
  fflush (stdout);
  ms = ms / 22;
  if (ms == 0)
    ms ++;
  for (i=0; i<ms; i++)
  {
#ifdef WIN32
    Sleep (1);
#else
//    usleep(0);
    usleep(1000);
#endif
    while (gtk_events_pending ())
      gtk_main_iteration ();
  }
}

/* gtk_isprint */
/**************************************************************************/
int gtk_isprint (            /* check if character is in printable range  */
  int c)                     /* the character                             */
/***************************************************************************
* globals     :                                                            *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Check whether a character iswithin printable range for GTK.              *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* 0 .. not printable                                                       *
* 1 .. is printable                                                        *
***************************************************************************/
{
  return 0x20<=c && c<=0x7e;
}
