/*-!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!
! Filename: gtk_multibox.c   Function: A Message Box object for GTK        !
! Part of:                                                                 !
!                                                                          !
!                                                                          !
! Vers.   Date    Auth.   Changes                                          !
! ------------------------------------------------------------------------ !
! 0.1    3.Mar.05  JH     Created                                          !
!      13.Oct.2005 JH     renamed into gtk_multibox                        !
!      15.Oct.2005 JH     the multibox is modal and transient              !
!      22.Oct.2005 JH     changed policy in the vbox (button fixed height) !
!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!-*/


#include <gtk/gtk.h>
#include <stdarg.h>
#include <stdlib.h>

#include "gtk_multibox.h"


/******************************  Defines  *********************************/

/******************************  Typedefs *********************************/

/********************************  Macros  ********************************/
#define debug(x) /*printf x*/

/**************************  Private Functions  ***************************/
static void multibox_close ( /* destroy the specified widget              */
  GtkWidget *parent,         /* the parent widget                         */
  gpointer  data             /* the widget to destroy                     */
);

/******************************  Globals  *********************************/
static GtkWidget *multibox;  /* the message box                           */
static int button_index=0;   /* return value                              */
static int *return_values;   /* values for the callback function          */
static int was_here=0;

/* gtk_multi_box */
/**************************************************************************/
int gtk_multi_box (          /* the multi message box main function       */
  GtkWidget *parent,         /* the parent window                         */
  char *title,               /* the title of the message box              */
  char *text,                /* the text to be displayed in the msg box   */
  int  no_of_buttons,        /* number of buttons                         */
  ...)                       /* label(s) of button(s) if any              */
/***************************************************************************
* globals     : multibox, button_index, return_values[]                    *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* This is the Message Box (with multiple buttons) object main function.    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* index of clicked button                                                  *
***************************************************************************/
{
  GtkWidget *label;
  GtkWidget *overall_box;
  GtkWidget *button_box;
  GtkWidget *button;
  va_list   ap;
  char      *button_label;
  int       i;

  was_here = 0;
  /* create the message box */
  multibox = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  /* we want it to be a modal window */
  gtk_window_set_modal (GTK_WINDOW(multibox), TRUE);
  /* ... and it should be on top of our main window */
  gtk_window_set_transient_for (GTK_WINDOW(multibox), GTK_WINDOW(parent));
  g_signal_connect (G_OBJECT (multibox), "destroy",
    G_CALLBACK (multibox_close), NULL);

  gtk_window_set_title (GTK_WINDOW (multibox), title);
  gtk_container_set_border_width (GTK_CONTAINER (multibox), 5);

  overall_box = gtk_vbox_new (FALSE, 10);
  gtk_container_add (GTK_CONTAINER (multibox), overall_box);
  gtk_container_border_width (GTK_CONTAINER (multibox), 10);

  /* put the text */
  label = gtk_label_new (text);
  gtk_box_pack_start_defaults (GTK_BOX(overall_box), label);

  /* do the buttons, if any */
  if (no_of_buttons)
  {
    /* create a hbox for the buttons */
    button_box = gtk_hbox_new (TRUE, 10);
    gtk_box_pack_start_defaults (GTK_BOX(overall_box), button_box);
    return_values = malloc (no_of_buttons * sizeof(*return_values));
    va_start (ap, no_of_buttons);
    for (i=0; i<no_of_buttons; i++)
    {
      return_values[i] = i;
      button_label = va_arg (ap, char *);
      button = gtk_button_new_with_label (button_label);
      g_signal_connect (G_OBJECT (button), "clicked",
        G_CALLBACK (multibox_close), (gpointer)(return_values[i]));
  
      gtk_box_pack_start_defaults (GTK_BOX(button_box), button);
      /* set the first button as the default button */
      if (i== 0)
        gtk_window_set_focus (GTK_WINDOW(multibox), button);
    }
    va_end (ap);
  }

  gtk_window_set_modal (GTK_WINDOW (multibox), TRUE);

  gtk_window_set_position(GTK_WINDOW(multibox),GTK_WIN_POS_CENTER_ON_PARENT);

  gtk_widget_show_all (multibox);

  gtk_main();

  return button_index;
}

/* multibox_close */
/**************************************************************************/
void multibox_close (        /* destroy the specified widget              */
  GtkWidget *parent,         /* the parent widget                         */
  gpointer  data)            /* the widget to destroy                     */
/***************************************************************************
* globals     : msgbox, button_index, return_values[]                      *
* called from : callback function                                          *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the specified widget and set the return value according to the   *
* clicked button.                                                          *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  /* for some reason this callback is always called twice */
  if (!was_here)
  {
    was_here = 1;
    button_index = (int)data;
    free (return_values);
    debug (("multibox_close(): button_index=%i\n", button_index));
    gtk_widget_destroy (multibox);
    gtk_main_quit ();
  }
}

