/*-!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!
! Filename: statusbox.c   Function: Status box, progress bar objects       !
! Part of:  jtag                                                           !
!                                                                          !
!                                                                          !
! Vers.   Date    Auth.   Changes                                          !
! ------------------------------------------------------------------------ !
! 0.5    6.Mar.05  JH     Added to GTK JTAG project                        !
!     15.Oct.2005  JH     statusbox is modal and transient                 !
!     19.Oct.2005  JH     set widget pointer to NULL after destroying      !
!     27.Oct.2006  JH     dummy callback function as static                !
!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!--!-*/

#include <gtk/gtk.h>         /* GTK stuff */

#include "gtk_statusbox.h"


/******************************  Defines  *********************************/
#define debug(x)  /* g_print x */

/**************************  Private Functions  ***************************/
static void dummy (GtkWidget *widget, gpointer data) {}

/*************************  Function Prototypes  **************************/

/******************************  Globals  *********************************/
static GtkWidget *progressbar;    /* the progress bar object              */


/* gtk_status_box */
/**************************************************************************/
GtkWidget *gtk_status_box (  /* the user callable status_box() function   */
  GtkWidget *parent,         /* the parent window                         */
  char      *title,          /* the title of the status box               */
  char      *text,           /* the text to be displayed in the status box*/
  unsigned  has_progressbar) /* whether to create a progress bar          */
/***************************************************************************
* globals     : progressbar                                                *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* This is the Status Box object main function.                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* index of clicked button                                                  *
***************************************************************************/
{
  GtkWidget *window;
  GtkWidget *label;
  GtkWidget *overall_box;
  GtkAdjustment *adj;

  /* create the status box */
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  /* we want it to be a modal window */
  gtk_window_set_modal (GTK_WINDOW(window), TRUE);
  /* ... and it should be on top of our main window */
  gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));
  g_signal_connect (G_OBJECT(window),"delete_event",G_CALLBACK(dummy),NULL);
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_container_set_border_width (GTK_CONTAINER (window), 20);
  debug (("gtk_status_box() 1: window=0x%8.8x\n", (unsigned)window));

  overall_box = gtk_vbox_new (FALSE, 10);
  gtk_container_add (GTK_CONTAINER (window), overall_box);
  gtk_container_border_width (GTK_CONTAINER (window), 10);

  /* put the text */
  label = gtk_label_new (text);
//  gtk_label_set_justify (GTK_LABEL(label), jtype);
  gtk_box_pack_start (GTK_BOX (overall_box), label, TRUE, FALSE, 0);
  debug (("gtk_status_box() 2: label=0x%8.8x\n", (unsigned)label));

  /* create the progress bar */
  if (has_progressbar)
  {
    adj = (GtkAdjustment *)gtk_adjustment_new (0, 0, 100, 0, 0, 0);
    progressbar = gtk_progress_bar_new_with_adjustment (adj);
    gtk_progress_set_format_string (GTK_PROGRESS (progressbar), "%p%%");    
    gtk_progress_set_show_text (GTK_PROGRESS (progressbar), 1);
    gtk_box_pack_start (GTK_BOX (overall_box), progressbar, TRUE, TRUE, 0);
    debug (("gtk_status_box() 3: bar=0x%8.8x\n", (unsigned)progressbar));
  } else
    progressbar = NULL;

  gtk_window_set_modal (GTK_WINDOW (window), TRUE);

  gtk_window_set_position (GTK_WINDOW(window),GTK_WIN_POS_CENTER_ON_PARENT);

  gtk_widget_show_all (window);

  return window;
}

/* gtk_status_box_destroy */
/**************************************************************************/
void gtk_status_box_destroy( /* destroy the status box                    */
  GtkWidget *widget)         /* the widget to destroy                     */
/***************************************************************************
* globals     : progressbar                                                *
* called from : callback function                                          *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the specified widget.                                            *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  debug (("gtk_status_box_destroy() 1: widget=0x%8.8x\n", (unsigned)widget));

  if (progressbar)
  {
    gtk_widget_destroy (GTK_WIDGET(progressbar));
    progressbar = NULL;
  }

  gtk_widget_destroy (GTK_WIDGET(widget));
  widget = NULL;
}

/* gtk_status_box_set_value */
/**************************************************************************/
void gtk_status_box_set_value (   /* set a new value for the progress bar */
  GtkWidget *statusbox,      /* parent of the progress bar                */
  int val)                   /* new value                                 */
/***************************************************************************
* globals     : progressbar                                                *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Set a new value to the progress bar, if any.                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_progress_set_value (GTK_PROGRESS (progressbar), val);
}

/* gtk_status_box_get_value */
/**************************************************************************/
int gtk_status_box_get_value (    /* get the curr. val of the progress bar*/
  GtkWidget *statusbox)      /* parent of the progress bar                */
/***************************************************************************
* globals     : progressbar                                                *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Set a new value to the progress bar, if any.                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  return gtk_progress_get_value (GTK_PROGRESS (progressbar));
}

/* gtk_status_box_has_progressbar */
/**************************************************************************/
int gtk_status_box_has_progressbar (   /* check for progress bar          */
  GtkWidget *statusbox)      /* the status box                            */
/***************************************************************************
* globals     : progressbar                                                *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Check whether the status box object has a progress bar attached to.      *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. no                                                                  *
* 1 .. yes                                                                 *
***************************************************************************/
{
  return progressbar ? 1 : 0;
}
