/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: io_ports.c
 *
 * Function: the I/O Ports Display window
 *
 * Part of:  DIY Calculator loader application
 *
 *
 * Changelog
 *
 *  0.1   6.Nov.2006   JH   new
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>

#ifndef WIN32
#include <unistd.h>          /* usleep() */
#endif

#include <string.h>          /* strchr() */
#include <stdlib.h>          /* atoi() */

#include "diy_load.h"
#include "io_ports.h"
#include "cpu_clk.h"
#include "cmd.h"
#include "serial.h"

#include "gtk_helper.h"
#include "gtk_statusbox.h"


/******************************  Defines  *********************************/


/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/**************************  Private Functions  ***************************/
static int read_reg (        /* read one register form the DIY Calculator */
  unsigned short addr,       /* the address to read                       */
  int *error                 /* error flag                                */
);


/************************  Function Prototypes  ***************************/


/******************************  Globals  *********************************/
static struct IO_PORTS io_ports;  /* the cpu register entry boxes         */
static GtkWidget *window;         /* the CPU register display window      */

extern int debug;
extern int offline;
extern int do_timer_cb;


/* io_window_refresh */
/**************************************************************************/
void io_window_refresh (     /* refresh the I/O Ports Display window      */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     :                                                            *
* called from : on clicking on the Refresh button                          *
* calls       : fill_io_ports()                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Refresh the I/O Ports display window.                                    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  fill_io_ports ();
}

/* io_window_destroy */
/**************************************************************************/
void io_window_destroy (     /* destroy the I/O Ports Display window      */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the I/O Ports display window                    *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the I/O Ports display window.                                    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* do_disp_io */
/**************************************************************************/
GtkWidget *do_disp_io (      /* the I/O Ports Display window              */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : windows, io_ports                                          *
* called from : disp_io()                                                  *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Create the I/O ports display window.                                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the I/O ports display window                                             *
***************************************************************************/
{
  GtkWidget *vbox, *hbox, *vbox1, *vbox2, *button_box;
  GtkWidget *frame, *box, *entry, *button;

  DEBUG (("do_disp_io()\n"));
  if (window == NULL)
  {
    /***************  create the "I/O Port Display" window  ***************/
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    /* install the delete_event handler that immediately exits GTK */
    g_signal_connect (G_OBJECT (window), "delete_event",
      G_CALLBACK (io_window_destroy), NULL);
    /* set the border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (window), 10);
    /* set the window title */
    gtk_window_set_title (GTK_WINDOW (window), "I/O Ports Display");
    /* when the main window will be closed, close also this one */
    gtk_window_set_destroy_with_parent (GTK_WINDOW(window), TRUE);
    /* set it on top of the main window */
    gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));

    /******************   create the various widgets  *********************/
    vbox = gtk_vbox_new (FALSE, 4);
    /* put the box into the window */
    gtk_container_add (GTK_CONTAINER (window), vbox);

    hbox = gtk_hbox_new (TRUE, 10);
    gtk_container_add (GTK_CONTAINER (vbox), hbox);

    button_box = gtk_hbox_new (TRUE, 10);
    gtk_container_add (GTK_CONTAINER (vbox), button_box);

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (hbox), vbox1);

    vbox2 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (hbox), vbox2);

    /* LCD */
    /* the frame with title */
    frame = gtk_frame_new ("Main Display $F031");
    /* center the title horizontally */
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    /* put it into the left vbox */
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    /* a box within the frame (for spacing) */
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    /* the text entry */
    entry = gtk_entry_new ();
    /* read only */
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    /* default text */
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    /* center alignment */
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    /* make it a decent size */
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    /* put it into the box */
    gtk_container_add (GTK_CONTAINER (box), entry);
    /* save it in the IO_PORTS object */
    io_ports.disp = entry;

    /* user LEDs */
    frame = gtk_frame_new ("LEDs $F032");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.led = entry;

    /* KeyPad */
    frame = gtk_frame_new ("KeyPad $F011");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.key = entry;

    /* IRQ source */
    frame = gtk_frame_new ("IRQ Source $F012");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox1), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.irq = entry;

    /* UART */
    frame = gtk_frame_new ("UART $F014/15");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    /* status register */
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.uart_s = entry;
    /* data register */
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.uart_d = entry;

    /* CPU clock */
    frame = gtk_frame_new ("CPU Clock $F036");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.clk = entry;

    /* SPI controller */
    frame = gtk_frame_new ("SPI Controller $F017/18");
    gtk_frame_set_label_align (GTK_FRAME(frame), 0.5, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 2);
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (box), 4);
    gtk_container_add (GTK_CONTAINER (frame), box);
    /* status register */
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.spi_s = entry;
    /* data register */
    entry = gtk_entry_new ();
    gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
    gtk_entry_set_text (GTK_ENTRY(entry), "$XX");
    gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
    gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
    gtk_container_add (GTK_CONTAINER (box), entry);
    io_ports.spi_d = entry;

    /* 'refresh' button */
    button = gtk_button_new_with_label ("Refresh");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(io_window_refresh), NULL);
    gtk_box_pack_start (GTK_BOX (button_box), button, TRUE, TRUE, 0);

    /* 'close' button */
    button = gtk_button_new_with_label ("Close");
    g_signal_connect (G_OBJECT(button), "clicked",
      G_CALLBACK(io_window_destroy), NULL);
    gtk_box_pack_start (GTK_BOX (button_box), button, TRUE, TRUE, 0);

    gtk_window_set_position (GTK_WINDOW (window), GTK_WIN_POS_CENTER_ON_PARENT);

    /* show the window */
    gtk_widget_show_all (window);
  }  /* have to create the IO ports display window */

  return window;
}

/* fill_io_ports */
/**************************************************************************/
int fill_io_ports ()         /* fill the I/O ports display window         */
/***************************************************************************
* globals     : io_ports, offline                                          *
* called from : disp_io(), io_window_refresh()                             *
* calls       : read_reg(), gtk_status_box(), gtk_status_box_set_value(),  *
*               gtk_status_box_destroy(), gtk_sleep()                      *
* Function                                                                 *
* --------                                                                 *
* Fill the CPU register display window with values.                        *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* 1 .. timeout                                                             *
***************************************************************************/
{
  char buf[BUFLEN], buf2[10];
  unsigned char reg;
  static unsigned char old_key=0xff;
  int error, all_error, freq, value;
  GtkWidget *statusbox;

  DEBUG (("fill_io_ports()\n"));

  value = 0;
  statusbox = gtk_status_box (window, "I/O Ports Display",
    "Receiving I/O ports data ... ", 1);
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (100);
  all_error = 0;

  /* LCD */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.disp), "offline");
  else
  {
    reg = read_reg (0xf031, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("Display   = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.disp), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* LEDs */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.led), "offline");
  else
  {
    reg = read_reg (0xf032, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("user LEDs = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.led), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* KeyPad */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.key), "offline");
  else
  {
    reg = read_reg (0xf011, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X/$%2.2X", reg, old_key);
      old_key = reg;
      DEBUG (("KeyPad    = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.key), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* IRQ source */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.irq), "offline");
  else
  {
    reg = read_reg (0xf012, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("IRQ src   = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.irq), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* UART status */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.uart_s), "offline");
  else
  {
    reg = read_reg (0xf014, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("UART stat = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.uart_s), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* UART data */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.uart_d), "offline");
  else
  {
    reg = read_reg (0xf015, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("UART data = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.uart_d), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* CPU clock */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.clk), "offline");
  else
  {
    reg = read_reg (0xf036, &error);
    if (!error)
    {
      freq = divider_to_freq (reg, buf2);
      sprintf (buf, "$%2.2X (%s)", reg, buf2);
      DEBUG (("CPU clock = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.clk), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* SPI controller status */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.spi_s), "offline");
  else
  {
    reg = read_reg (0xf017, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("SPI stat  = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.spi_s), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);

  /* SPI controller data */
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(io_ports.spi_d), "offline");
  else
  {
    reg = read_reg (0xf018, &error);
    if (!error)
    {
      sprintf (buf, "$%2.2X", reg);
      DEBUG (("SPI data  = %s\n", buf));
      gtk_entry_set_text (GTK_ENTRY(io_ports.spi_d), buf);
    } else
      all_error ++;
  }
  value += 14;
  gtk_status_box_set_value (statusbox, value);
  gtk_sleep (1);
  gtk_status_box_destroy (statusbox);

  return all_error;
}

/* read_reg */
/**************************************************************************/
int read_reg (               /* read one register form the DIY Calculator */
  unsigned short addr,       /* the address to read                       */
  int *error)                /* error flag                                */
/***************************************************************************
* globals     : io_ports, do_timer_cb                                      *
* called from : fill_io_ports()                                            *
* calls       : write_port(), read_line(), wait_for_prompt()               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Read one register from the DIY Calculator CPU.                           *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* 1 .. timeout                                                             *
***************************************************************************/
{
  char cmd[20], buf[BUFLEN];
  unsigned char c;

  *error = 0;
  sprintf (cmd, "md %x 1\r\n", addr);
  do_timer_cb = 0;
  write_port (cmd, strlen(cmd));
  usleep (200000);

  /* command line */
  read_line (buf, BUFLEN);

  /* empty line */
  read_line (buf, BUFLEN);

  /* line "Memory Display ... */
  read_line (buf, BUFLEN);

  /* line with interesting data */
  read_line (buf, BUFLEN);

  buf[8] = '\0';
  do_timer_cb = 1;
  c = atox (buf+6);
  *error = wait_for_prompt ();
//  DEBUG (("read_reg(): c=0x%2.2x, error=%d\n", c, *error));
  return c;
}
