/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: main.c
 *
 * Function: main module of loader application for DIY Calculator
 *
 * Modules:
 *  background.c ..... the DIY Calculator background image
 *  cmd.c ............ send a cmd to the DIY Calculator, wait for a string
 *  cpu_clk.c ........ adjust CPU speed dialog
 *  cpu_reg.c ........ display DPU registers
 *  disp_mem ......... the Memory Walker
 *  io_ports.c ....... display some well-known I/O ports
 *  save_ram.c ....... save the contents of the RAM to Flash
 *  serial.c ......... low level serial I/O functions
 *  set_comm_port.c .. select the communications port dialog
 *  version_info.c ... display DIY Calculator version information
 *  gtk_filename.c ... file selector dialog
 *  gtk_helper.c ..... helper functions for GTK
 *  gtk_msgbox.c ..... a simple message box
 *  gtk_multibx.c .... a message box with multiple buttons
 *  gtk_statusbox.c .. a status box with or without progress bar
 *
 *
 * Changelog
 *
 *  0.1  10.Oct.2006   JH   new
 *  0.2  23.Oct.2006   JH   
 *  0.9  24.Oct.2006   JH   new design
 *  1.0  10.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */


#include <gtk/gtk.h>

#ifdef WIN32
#include <fcntl.h>           /* _O_TEXT */
#include <conio.h>           /* getch() */
#else
#include <unistd.h>          /* usleep() */
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>           /* tolower() */

#include "diy_load.h"
#include "cpu_reg.h"
#include "io_ports.h"
#include "cpu_clk.h"
#include "save_ram.h"
#include "disp_mem.h"
#include "serial.h"
#include "cmd.h"
#include "set_comm_port.h"
#include "version_info.h"

#include "gtk_helper.h"
#include "gtk_msgbox.h"
#include "gtk_statusbox.h"
#include "gtk_filename.h"


/******************************  Defines  *********************************/
#ifdef WIN32
#define DEFAULT_PORT "COM1"
#else
#define DEFAULT_PORT "/dev/ttyS0"
#endif

/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/****************************  Private Functions  *************************/
static void create_widgets ( /* create all widgets we need                */
  GtkWidget *window,         /* the main window                           */
  int do_viewer              /* whether to do the viewer or backgnd image */
);

static void add_window_title ( /* add a string to the active window title */
  GtkWidget *window,         /* the window                                */
  char *s                    /* the string                                */
);

static void usage (void);    /* print usage inforamtion and exit          */

static void printf_statusline (   /* print formatted text to the stausline*/
  char *fmt,                 /* format string                             */
  ...                        /* arguments                                 */
);

static void set_mode (       /* set a new mode and update status line     */
  int new_mode               /* the new mode: (dis-) connected, step, etc.*/
);

/****************************  Callback Functions  ************************/
static void version_info (void);
static void do_connect (void);
static void exit_program (void);
static void set_cpu_clock (void);
static void set_comm_port (void);
static void disp_mem (void);
static void disp_cpu_reg (void);
static void disp_io (void);
static void load_ram (void);
static void save_ram (void);
static void purge_ram (void);
static void exec_term (void);
static void help_about (void);
static void do_reset (void);
static void do_step (void);
static void do_run (void);
static void viewer_callback (void);
static int timer_callback (gpointer data);


/******************************  Globals  *********************************/
GtkWidget *main_window;      /* the main window                           */
GtkWidget *cpu_reg_window;   /* the CPU Register display window           */
GtkWidget *io_window;        /* the I/O Ports display window              */
GtkWidget *cpu_clk_window;   /* the CPU clock speed adjustment dialog box */
GtkWidget *version_window;   /* the Version info display window           */
GtkWidget *save_ram_window;  /* the Save RAM dialog box                   */
GtkWidget *terminal_window;  /* the terminal window                       */
GtkItemFactory *item_factory;   /* the menu entries of the main window    */
GtkWidget *mem_window;       /* the memory display window                 */
GtkWidget *overall_box;      /* main box in main window                   */
GtkWidget *viewer;           /* the terminal window text box              */
GtkWidget *statusline;       /* status line                               */
int portfd;                  /* the communications port file descriptor   */
int mode;                    /* (dis)connected, step, run, etc.           */
int timer;                   /* timer object for terminal srceen update   */
int do_timer_cb;             /* if 0 the timer callback does nothing      */
int is_terminal;             /* if 1 the terminal is currently active     */

int start_addr;              /* start addr for Memory Display and Save RAM*/
int end_addr;                /* end addr for Memory Display and Save RAM  */

extern const guint8 background_inline[]; /* bg image data: background.c   */

/* the menus and entries */
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",                  NULL,         NULL,           0, "<Branch>"},
  { "/File/_Connect",          NULL,         do_connect,     0, "<Item>"},
  { "/File/sep1",              NULL,         NULL,           0, "<Separator>"},
  { "/File/_Version Info",     "<ALT>V",     version_info,   0, "<Item>"},
  { "/File/sep2",              NULL,         NULL,           0, "<Separator>"},
  { "/File/E_xit",             "<ALT>Q",     exit_program,   0, "<Item>"},
  { "/Se_tup",                 NULL,         NULL,           0, "<Branch>"},
  { "/Setup/System _Clock",    "<ALT>C",     set_cpu_clock,  0, "<Item>"},
  { "/Setup/C_ommunications Port","<ALT>O",  set_comm_port,  0, "<Item>" },
  { "/_Display",               NULL,         NULL,           0, "<Branch>"},
  { "/Display/Memory _Walker", "<ALT>W",     disp_mem,       0, "<Item>"},
  { "/Display/_CPU Registers", "<ALT>R",     disp_cpu_reg,   0, "<Item>"},
  { "/Display/_IO Ports",      "<ALT>I",     disp_io,        0, "<Item>"},
  { "/_Memory",                NULL,         NULL,           0, "<Branch>"},
  { "/Memory/_Load RAM",       "<ALT>L",     load_ram,       0, "<Item>"},
  { "/Memory/_Save to Flash",  "<ALT>S",     save_ram,       0, "<Item>"},
  { "/Memory/_Purge RAM",      "<ALT>P",     purge_ram,      0, "<Item>"},
  { "/_Tools",                 NULL,         NULL,           0, "<Branch>"},
  { "/Tools/T_erminal",        "<ALT>E",     exec_term,      0, "<Item>"},
  { "/_Help",                  NULL,         NULL,           0, "<LastBranch>"},
  { "/Help/_About",            NULL,         help_about,     0, "<Item>"},
};

static int nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

struct COMM_PORT portlist[MAX_COMM_PORTS];/* list of available comm. ports*/
int no_of_ports;             /* number of available comm. pports          */

/* command line options */
char portname[20];           /* communications port name                  */
int debug = 0;               /* swith on debug messages                   */
int terminal = 0;            /* run terminal only                         */
int offline = 0;             /* no DIY Calculcator HW connected           */

enum MODES
{
  M_DISCONNECTED,
  M_CONNECTED,
  M_RESET,
  M_SSTEP,
  M_RUN,
  M_OFFLINE
};

static char *mode_names[] =
{
  "Not connected",
  "Connected",
  "Reset",
  "Single Step",
  "Running",
  "Offline"
};

/* get rid of the console window in Windows, be sure to link as Windows GUI
 * application (-mwindows option to the linker)
 */
#ifdef WIN32
int main (int argc, char *argv[]);

int WINAPI WinMain (
  HINSTANCE hInstance,
  HINSTANCE hPrevInstance,
  LPSTR lpCmdLine,
  int nCmdShow)
{
  return main (__argc, __argv);
}
#endif

/* main */
/**************************************************************************/
int main (                   /*                                           */
  int argc,                  /*                                           */
  char *argv[])              /*                                           */
/***************************************************************************
* globals     : main_window, portfd, portlist, no_of_ports, portname, mode *
*               offline, terminal, do_timer_cb, is_terminal, start_addr,   *
*               debug, end_addr                                            *
* called from :                                                            *
* calls       : create_widgets(), add_window_title(), usage(), init_port() *
*               get_avail_ports(), exec_term(), close_port()               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* - initialise GTK                                                         *
* - parse command line                                                     *
* - find available communications ports                                    *
* - open selected comm. port                                               *
* - create main window                                                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* 1 .. error                                                               *
***************************************************************************/
{
  GtkWidget *window;
  int i;
  char opt;
#ifdef WIN32
  int fdout;
  FILE *out;
#endif

  /* initialise GTK stuff */
  gtk_init (&argc, &argv);

  /* default values */
  debug = 0;
  strcpy (portname, DEFAULT_PORT);
  terminal = 0;
  offline = 0;

  mode = M_DISCONNECTED;
  timer = 0;
  do_timer_cb = 1;
  is_terminal = 0;
  start_addr = 0x4000;
  end_addr = 0x40ff;

  /* parse cmd line */
  for (i=1; i<argc; i++)
  {
    if (*argv[i] == '-' || *argv[i] == '/')
    {
      argv[i] ++;
      if (*argv[i] == '-' || *argv[i] == '/')
      {
        /* is a long option */
        argv[i] ++;
        if (strcmp (argv[i], "debug") == 0)
          opt = 'd';
        else if (strcmp (argv[i], "offline") == 0)
          opt = 'o';
        else if (strncmp (argv[i], "portfd", 4) == 0)
          opt = 'p';
        else if (strcmp (argv[i], "terminal") == 0)
          opt = 't';
        else if (strncmp (argv[i], "version", 7) == 0)
          opt = 'v';
        else
          opt = 'h';

        while (isalpha (argv[i][0]) || argv[i][0] == '-')
          argv[i] ++;
      } else
      {
        /* short option */
        opt = argv[i][0];
        argv[i] ++;
      }
      if (argv[i][0] == '=')
        argv[i] ++;
      switch  (opt)
      {
        case 'd': /* debug */
#ifdef WIN32
          /* create a console for stdio output */
          AllocConsole ();
          fdout = _open_osfhandle
                    ((long)GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
          out = fdopen (fdout, "w");
          *stdout = *out;
#endif
          debug = 1;
          break;
        case 'o': /* offline mode */
          offline = 1;
          break;
        case 'p': /* communications port name */
          strcpy (portname, argv[i]);
#ifdef WIN32
          strupr (portname);
#endif
          break;
        case 't': /* terminal mode */
          terminal = 1;
          break;
        case 'v': /* version information */
#ifdef WIN32
          /* create a console for stdio output */
          AllocConsole ();
          fdout = _open_osfhandle
                    ((long)GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
          out = fdopen (fdout, "w");
          *stdout = *out;
#endif
          printf ("\n%s V%s %s\n", PROGNAME, VER, __DATE__);
#ifdef WIN32
          /* close the console */
#ifdef RELEASE_VERSION
          printf ("\npress any key ... ");
          getch();
#endif
          FreeConsole ();
#endif
          return 0;;
        case 'h': /* cmdline help */
        default:
          usage ();
      } /* switch (opt) */
    } /* is switch char */
  } /* for loop */

  DEBUG (("cmdline options:\n"));
  DEBUG (("  debug=%d\n", debug));
  DEBUG (("  terminal=%d\n", terminal));
  DEBUG (("  portname=%s\n", portname));
  DEBUG (("  offline=%d\n", offline));
  DEBUG (("\n"));

  /* create the main window */
  main_window = window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

  /* install the delete_event handler that immediately exits GTK */
  g_signal_connect (G_OBJECT (window), "delete_event",
                      G_CALLBACK (gtk_main_quit), NULL);

  /* set the border width of the window */
  gtk_container_set_border_width (GTK_CONTAINER (window), 10);

  /* create all widgets we want to use */
  create_widgets (window, 0);

  /* add the program name to the window title bar */
  add_window_title (window, "");

  /* show the main window */
  gtk_widget_show (window);

  /* detect available communcations ports */
  no_of_ports = get_avail_ports ();

  /* open the communications port */
  portfd = init_port (portname);

  /* start the terminal */
  if (terminal)
    exec_term ();

  /* enter the event loop */
  gtk_main ();

  close_port (portfd);
  return 0;
}

/***************************  Callback Functions  *************************/
/* do_connect */
/**************************************************************************/
void do_connect ()           /* connect to the DIY Calculator             */
/***************************************************************************
* globals     : main_window, offline, do_timer_cb                          *
* called from : menu "/File/Connect"                                       *
* calls       : write_port(), set_mode(), gtk_msg_box(), wait_for_prompt() *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Connect to the DIY Calculator                                            *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char cmd[] = "xy\n";

  DEBUG (("do_connect() ... "));

  if (offline)
  {
    DEBUG (("offline\n"));
    set_mode (M_OFFLINE);
  } else
  {
    do_timer_cb = 0;
    write_port (cmd, strlen (cmd));
    if (wait_for_prompt() == OK)
    {
      set_mode (M_CONNECTED);
      DEBUG (("OK\n"));
    } else
    {
      DEBUG (("timeout\n"));
      set_mode (M_DISCONNECTED);
      gtk_msg_box (main_window,
        "ERROR do_connect()", "Cannot connect to DIY Calculator!", 0);
    }
    do_timer_cb = 1;
  }
}

/* version_info */
/**************************************************************************/
void version_info ()         /* display DIY Calculator versions           */
/***************************************************************************
* globals     : main_window, mode                                          *
* called from : menu "/File/Version Info"                                  *
* calls       : do_connect(), do_version_info()                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display DIY Calculator versions                                          *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("version_info()\n"));
  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
    version_window = do_version_info (main_window);
}

/* exit_program */
/**************************************************************************/
void exit_program ()         /* single point of exit                      */
/***************************************************************************
* globals     : timer                                                      *
* called from : menu "/File/Exit"                                          *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* A single point of exit.                                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("exit_program()\n"));
  if (timer)
    gtk_timeout_remove (timer);

  gtk_main_quit();
}

/* set_cpu_clock */
/**************************************************************************/
void set_cpu_clock ()        /* set a new CPU clock frequency             */
/***************************************************************************
* globals     : main_window, mode, cpu_clk_window                          *
* called from : menu "/Setup/System Clock"                                 *
* calls       : do_connect(), do_cpu_clk()                                 *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Set a new CPU clock frequency.                                           *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("set_cpu_clock()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
    cpu_clk_window = do_cpu_clk (main_window);
}

/* set_comm_port */
/**************************************************************************/
void set_comm_port ()        /* define the used communications port       */
/***************************************************************************
* globals     : main_window, portfd                                        *
* called from : menu "/Setup/Set Communications Port"                      *
* calls       : do_set_comm_port()                                         *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Define the used communications port.                                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("set_comm_port()\n"));
  portfd = do_set_comm_port (main_window);
}

/* disp_mem */
/**************************************************************************/
void disp_mem ()             /* display a memory region                   */
/***************************************************************************
* globals     : main_window, mode, mem_window                              *
* called from : menu "/Display/Memory Walker"                              *
* calls       : do_connect(), set_mode(), do_disp_mem()                    *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display a memory region.                                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("disp_mem()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    mem_window = do_disp_mem (main_window);
    if (mem_window == NULL)
      set_mode (M_DISCONNECTED);
  }
}

/* disp_cpu_reg */
/**************************************************************************/
void disp_cpu_reg ()         /* display the CPU registers                 */
/***************************************************************************
* globals     : main_window, mode, offline, do_timer_cb, cpu_reg_window    *
* called from : menu "/Display CPU Registers"                              *
* calls       : do_connect(), do_disp_cpu_reg(), send_cmd(), set_mode()    *
*               read_line(), fill_cpu_regs()                               *
* Function                                                                 *
* --------                                                                 *
* Display the contents of the CPU registers.                               *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char buf[BUFLEN];

  DEBUG (("disp_cpu_reg()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    cpu_reg_window = do_disp_cpu_reg (main_window);
    if (cpu_reg_window == NULL)
      set_mode (M_DISCONNECTED);
    else
    {
      if (offline)
        fill_cpu_regs ();
      else
      {
        send_cmd (DUMPREG);
        do_timer_cb = 0;
        usleep(300000);
  
        /* wait for title line */
        do
          read_line (buf, BUFLEN);
        while (strstr(buf, "register dump") == NULL);
  
        if (fill_cpu_regs () != OK)
          set_mode (M_DISCONNECTED);
        do_timer_cb = 1;
      }
    }
  }
}

/* disp_io */
/**************************************************************************/
void disp_io ()              /* display well-known I/O ports              */
/***************************************************************************
* globals     : main_window, mode, io_window                               *
* called from : menu "/Display/IO Ports"                                   *
* calls       : do_connect(), set_mode(), do_disp_io(), fill_io_ports()    *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the well-known I/O port addresses.                               *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("disp_io()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    io_window = do_disp_io (main_window);
    if (io_window == NULL)
      set_mode (M_DISCONNECTED);
    else
    {
      if (fill_io_ports () != OK)
        set_mode (M_DISCONNECTED);
    }
  }
}

/* load_ram */
/**************************************************************************/
void load_ram ()             /* upload a .ram file                        */
/***************************************************************************
* globals     : main_window, mode, do_timer_cb, offline                    *
* called from : menu "/Memory/Load RAM"                                    *
* calls       : do_connect(), send_cmd(), read_line(), write_port(),       *
*               gtk_status_box(), gtk_status_box_set_value(), gtk_sleep()  *
*               gtk_status_box_destroy(),gtk_msg_box(),printf_statusline() *
*               wait_for_string(), gtk_filename()                          *
* Function                                                                 *
* --------                                                                 *
* Upload a .ram file to the DIY Calculator.                                *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  FILE *file;
  char *filename;
  char buf[BUFLEN];
  int written;
  long file_len;
  int  i, lineno, no_of_lines, delta, value;
  GtkWidget *statusbox;

  DEBUG (("load_ram()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    filename = gtk_filename (main_window, FN_LOAD);
    if (filename)
    {
      file = fopen (filename, "rb");
      if (file)
      {
        /* now send the file to the DIY Calculator */
        do_timer_cb = 0;
        sprintf (buf, "Loading DIY Calculator with \"%s\", please wait ...",
          filename);
        if (offline)
          statusbox = gtk_status_box(main_window,PROGNAME" - offline",buf,1);
        else
          statusbox = gtk_status_box (main_window, PROGNAME, buf, 1);
        value = 0;
        gtk_status_box_set_value (statusbox, value);
        gtk_sleep (1);
        fseek (file, 0, SEEK_END);
        file_len = ftell (file);
        rewind (file);
        if (!offline)
        {
          send_cmd (LOAD);
          /* eat an additional line */
          read_line (buf, BUFLEN);
        }
        /* compute the number of lines:
         *   1st line: 47 chars  ----- # FILE TYPE ...
         *   2nd line: ca. 30 chars -- # GENERATED ...
         *   3rd line: 39 chars  ----- # DATE-TIME ...
         *   4th line: ca. 25 chars -- # SOURCEWAS ...
         *   5th line: 2 chars  ------ <empty line>
         *   6th line: 18 chars  ----- STARTADDR
         *   7th line: 18 chars  ----- FINALADDR
         *   8th line: 2 chars  ------ <empty line>
         *   data lines: 55 chars each
         */
        no_of_lines = (file_len - 181) / 55 + 6; /* dont count empty lines */
        if (no_of_lines == 0)
          no_of_lines ++;
        if (no_of_lines > 100)
          delta = no_of_lines / 100;
        else
          delta = 100 / no_of_lines;

        /* first write the four comment lines */
        lineno = 0;
#if 0
        for (i=0; i<4; i++)
        {
          fgets (buf, BUFLEN, file);
          if (feof(file))
            break;
//          DEBUG (("load_ram(): writing \"%s\" ... ", buf));
          if (!offline)
          {
            written = write_port (buf, strlen (buf));
//          DEBUG (("%d OK\n", written));
          }
          lineno ++;
          if (strlen(buf) < 5)
            continue;
          if ((no_of_lines>100 && (lineno%delta)==0) || no_of_lines<=100)
          {
            if (no_of_lines <= 100)
              value += delta;
            else
              value ++;
            if (value > 100)
              value = 100;
            gtk_status_box_set_value (statusbox, value);
          }
          gtk_sleep (100);
        }
        /* one empty line */
        fgets (buf, BUFLEN, file);
        if (!offline)
          write_port (buf, strlen(buf));
        /* write the start/end address */
        for (i=0; i<2; i++)
        {
          fgets (buf, BUFLEN, file);
          if (feof(file))
            break;
//          DEBUG (("load_ram(): writing \"%s\" ... ", buf));
          if (!offline)
          {
            written = write_port (buf, strlen (buf));
//          DEBUG (("%d OK\n", written));
          }
          lineno ++;
          if (strlen(buf) < 5)
            continue;
          if ((no_of_lines>100 && (lineno%delta)==0) || no_of_lines<=100)
          {
            if (no_of_lines <= 100)
              value += delta;
            else
              value ++;
            if (value > 100)
              value = 100;
            gtk_status_box_set_value (statusbox, value);
          }
          gtk_sleep (200);
        }
        /* one empty line */
        fgets (buf, BUFLEN, file);
        if (!offline)
          write_port (buf, strlen(buf));
#else
        for (i=0; i<6; i++)
        {
          fgets (buf, BUFLEN, file);
          if (feof(file))
            break;
//          DEBUG (("load_ram(): writing \"%s\" ... ", buf));
          if (!offline)
          {
            written = write_port (buf, strlen (buf));
//          DEBUG (("%d OK\n", written));
          }
          lineno ++;
          if (strlen(buf) < 5)
            continue;
          if ((no_of_lines>100 && (lineno%delta)==0) || no_of_lines<=100)
          {
            if (no_of_lines <= 100)
              value += delta;
            else
              value ++;
            if (value > 100)
              value = 100;
            gtk_status_box_set_value (statusbox, value);
          }
          gtk_sleep (100);
        }
#endif
        /* now do the data lines */
        while (1)
        {
          fgets (buf, BUFLEN, file);
          if (feof(file))
            break;
//          DEBUG (("load_ram(): writing \"%s\" ... ", buf));
          if (offline)
            written = strlen (buf);
          else
            written = write_port (buf, strlen (buf));
//          DEBUG (("%d OK\n", written));
          if ((no_of_lines>100 && (lineno%delta)==0) || no_of_lines<=100)
          {
            if (no_of_lines <= 100)
              value += delta;
            else
              value ++;
            if (value > 100)
              value = 100;

            gtk_status_box_set_value (statusbox, value);
          }
          gtk_sleep (100);
        }
        gtk_status_box_set_value (statusbox, 100);
        fclose (file);
        /* wait for the response */
        if (!offline && wait_for_string ("Success!") != OK)
        {
          gtk_status_box_destroy (statusbox);
          sprintf (buf, "Error loading file \"%s\"", filename);
          gtk_msg_box (main_window, "Error", buf, 0);
          printf_statusline ("Error loading \"%s\"", filename);
        } else
        {
          gtk_status_box_destroy (statusbox);
          statusbox = gtk_status_box (main_window, PROGNAME, "Success!", 1);
          gtk_sleep (10);
          gtk_status_box_set_value (statusbox, 100);
          gtk_sleep (10);
          usleep (300000);
          gtk_status_box_destroy (statusbox);
          printf_statusline ("Loaded \"%s\" successfully", filename);
        }
        do_timer_cb = 1;
      } else
      {
        sprintf (buf, "Cannot open file \"%s\"!", filename);
        gtk_msg_box (main_window, "Load File", buf, 0);
      }
    }
  }
}

/* save_ram */
/**************************************************************************/
void save_ram ()             /* save the user program to flash ROM        */
/***************************************************************************
* globals     : main_window, mode, save_ram_window                         *
* called from : menu "/Memory/Save RAM"                                    *
* calls       : do_connect(), do_save_ram()                                *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Save a memory range in RAM to the onboard PSI Flash.                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  DEBUG (("save_ram()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
    save_ram_window = do_save_ram (main_window);
}

/* purge_ram */
/**************************************************************************/
void purge_ram ()            /* fill the RAM with a default pattern       */
/***************************************************************************
* globals     : mode, offline                                              *
* called from : menu "/Memory/Purge RAM"                                   *
* calls       : do_connect(), write_port()                                 *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Fill the DIY Calculator's memory with a default pattern.                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char cmd[]="mw 0 4000 700\n";

  DEBUG (("purge_ram()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (!offline && mode >= M_CONNECTED)
    write_port (cmd, strlen(cmd));
}

/* exec_term */
/**************************************************************************/
void exec_term ()            /* display a terminal window                 */
/***************************************************************************
* globals     : main_window, mode, do_timer_cb, is_terminal, mode, viewer  *
*               overall_box, item_factory                                  *
* called from : menu "/Tools/Terminal"                                     *
* calls       : create_widgets(), write_port(), read_line(), set_mode()    *
*               wait_for_prompt(), gtk_text_view_printf(), gtk_msg_box()   *
* Function                                                                 *
* --------                                                                 *
* Start a terminal to the DIY Calculator.                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char cmd[] = "xy\n";
  char buf[BUFLEN];
  int old_mode;
  static GtkItemFactoryEntry close_terminal = {
    "/Tools/Clos_e Terminal", "<ALT>E", exec_term, 0, "<Item>"};

  DEBUG (("exec_term()\n"));

  old_mode = mode;
  do_timer_cb = 0;
  if (viewer == NULL)
  {
    /* create the terminal */
    gtk_widget_destroy (overall_box);
    create_widgets (main_window, 1);

    if (viewer == NULL)
    {
      DEBUG (("Cannot create terminal window!\n"));
    } else
    {
      if (offline)
        gtk_text_view_printf (viewer, "===  offline mode  ===\n");
      else
      {
        read_line (buf, BUFLEN);
        write_port (cmd, strlen (cmd));
        usleep (200000);
        if (wait_for_prompt() == OK)
        {
          set_mode (M_CONNECTED);
          DEBUG (("OK\n"));
        } else
        {
          DEBUG (("timeout\n"));
          set_mode (M_DISCONNECTED);
          gtk_msg_box (main_window,
            "ERROR exec_term()", "Cannot connect to DIY Calculator!", 0);
        }
      }
      gtk_text_view_printf (viewer, "# ");
      is_terminal = 1;
      gtk_item_factory_delete_item (item_factory, "/Tools/Terminal");
      gtk_item_factory_create_item (item_factory, &close_terminal, NULL, 2);
      gtk_widget_grab_focus (viewer);
    }
  } else
  {
    /* close the terminal and show the background picture again */
    is_terminal = 0;
    gtk_widget_destroy (overall_box);
    viewer = NULL;
    create_widgets (main_window, 0);
  }
  do_timer_cb = 1;
  set_mode (old_mode);
}

/* help_about */
/**************************************************************************/
void help_about ()           /* display the "about" message box           */
/***************************************************************************
* globals     : main_window, offline                                       *
* called from : menu "/Help/About"                                         *
* calls       : gtk_msg_box()                                              *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the "About" message box.                                         *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char title[200], text[400];

  DEBUG (("help_about()\n"));
  sprintf (title, "About %s", PROGNAME);
  sprintf (text,"%s V%s, %s%s",PROGNAME,VER,__DATE__,offline?" - offline":"");
  gtk_msg_box (main_window, title, text, 0);
}

/* do_reset */
/**************************************************************************/
void do_reset ()             /* reset the DIY Calculator                  */
/***************************************************************************
* globals     : main_window, offline, do_timer_cb                          *
* called from : "Reset" button                                             *
* calls       : write_port(), set_mode(), wait_for_prompt(), gtk_msg_box() *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* (Soft-) Reset the DIY Calculator.                                        *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char c;

  DEBUG (("do_reset() ... "));

  if (offline)
  {
    DEBUG (("OK\n"));
    set_mode (M_RESET);
  } else
  {
    c = CTRL_C;
    write_port (&c, 1);
    do_timer_cb = 0;
    if (wait_for_prompt() == OK)
    {
      DEBUG (("OK\n"));
      set_mode (M_RESET);
    } else
    {
      DEBUG (("timeout\n"));
      gtk_msg_box (main_window, "Error",
        "Could not soft-reset the DIY Calculator.\n \
Please press reset button on DIY Calculator!", 0);
      set_mode (M_DISCONNECTED);
    }
    do_timer_cb = 1;
  }
}

/* do_step */
/**************************************************************************/
void do_step ()              /* single-step the DIY Calculator user pgm   */
/***************************************************************************
* globals     : main_window, mode, cpu_reg_window, do_timer_cb, offline    *
* called from : "Step" button                                              *
* calls       : do_conect(), do_disp_cpu_reg(), write_port(), read_line()  *
*               set_mode(), fill_cpu_regs()                                *
* Function                                                                 *
* --------                                                                 *
* Single-step advance the DIY Calculator user program.                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char c;
  char buf[BUFLEN];

  DEBUG (("do_step()\n"));

  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    cpu_reg_window = do_disp_cpu_reg (main_window);
    if (cpu_reg_window == NULL)
      set_mode (M_DISCONNECTED);
    else
    {
      c = CTRL_A;
      if (!offline)
      {
        write_port (&c, 1);
        do_timer_cb = 0;
        usleep (200000);
        /* wait for title line */
        do
          read_line (buf, BUFLEN);
        while (strstr (buf, "register dump") == NULL);
      }
      set_mode (M_SSTEP);

      if (fill_cpu_regs () != OK)
        set_mode (M_DISCONNECTED);
      do_timer_cb = 1;
    }
  }
}

/* do_run */
/**************************************************************************/
void do_run ()               /* start the DIY Calculator user program     */
/***************************************************************************
* globals     : mode, offline                                              *
* called from : "Run" button                                               *
* calls       : do_connect(), write_port(), set_mode()                     *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Start (or continue) the DIY Calculator user prgram.                      *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char c;

  DEBUG (("do_run()\n"));
  if (mode < M_CONNECTED)
    do_connect ();

  if (mode >= M_CONNECTED)
  {
    c = CTRL_B;
    if (!offline)
      write_port (&c, 1);
    set_mode (M_RUN);
  }
}

/* viewer_callback */
/**************************************************************************/
void viewer_callback ()      /* get a character from the terminal         */
/***************************************************************************
* globals     : offline, viewer                                            *
* called from : input from terminal window                                 *
* calls       : write_port(), gtk_text_view_getch(), gtk_text_view_printf()*
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Get a character from the terminal window and send it to DIY calculator.  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  char c;

  DEBUG (("viewer_callback(): "));
  c = gtk_text_view_getch (viewer);
  if (isprint (c))
  {
    DEBUG (("%c\n", c));
  } else
  {
    DEBUG (("0x%2.2x\n", c));
  }
  if (c == '\0')
    c = '\n';

  if (offline)
  {
    gtk_text_view_printf (viewer, "%c", c);
    if (c == '\n')
      gtk_text_view_printf (viewer, "===  offline mode  ===\n# ");
  } else
    write_port (&c, 1);
}

/* timer_callback */
/**************************************************************************/
int timer_callback (         /* the callback function for the timer       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : main_window,do_timer_cb,is_terminal,viewer,cpu_reg_window  *
* called from : gets called ervery 20 msec                                 *
* calls       : set_mode(), read_port(), do_disp_cpu_reg(), fill_cpu_regs()*
*               do_disp_mem(), gtk_text_view_printf()                      *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* This is the callback function for the timer.                             *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
  char buf[BUFLEN];
  size_t i;

//  DEBUG (("timer_callback(): "));
  if (do_timer_cb)
  {
    i = read_port (buf, BUFLEN);
    if (i > 0)
    {
      buf[i] = '\0';
//      DEBUG (("timer_callback(): buf=\"%s\"\n", buf));
      if (is_terminal)
      {
        gtk_text_view_printf (viewer, buf);
      } else if (strstr(buf, "register dump"))
      {
        set_mode (M_SSTEP);
        cpu_reg_window = do_disp_cpu_reg (main_window);
        usleep (200000);
        if (cpu_reg_window)
          fill_cpu_regs ();
      } else if (strstr(buf, "Continuing") || strstr(buf, "Starting"))
      {
        set_mode (M_RUN);
      } else if (strstr(buf, "ROM monitor"))
      {
        set_mode (M_RESET);
        gtk_text_view_printf (viewer, buf);
      } else if (strstr(buf, "Memory display"))
      {
        do_disp_mem (main_window);
      } else
        gtk_text_view_printf (viewer, buf);
    }
  }
//  DEBUG (("timer_callback() exiting\n"));
  /* as this is a timeout function, return TRUE so that it
   * continues to get called */
  return TRUE;
}

/****************************  Private Functions  *************************/
/* create_widgets */
/**************************************************************************/
void create_widgets (        /* create all widgets we need                */
  GtkWidget *window,         /* the main window                           */
  int do_viewer)             /* whether to do the viewer or backgng image */
/***************************************************************************
* globals     : menu_items, nmenu_items, overall_box, item_factory, viewer *
*               timer, background_inline, statusline                       *
* called from : main(), exec_term()                                        *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Create the menus, text box, and buttons                                  *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  GtkWidget *button;
  GtkWidget *button_box;
  GtkWidget *menubar;
  GtkAccelGroup *accel_group;
  GtkWidget *text_box;
  GtkTextBuffer *buffer;
  PangoFontDescription *font;
//  GdkColor color;
  GdkPixbuf *backgnd;
  GtkWidget *bg_image;

  /* first create an overall box to arrange all other widgets */
  overall_box = gtk_vbox_new (FALSE, 0);

  /* put the box into the main window */
  gtk_container_add (GTK_CONTAINER (window), overall_box);

  /******************************  the menu  ****************************/
  /* Make an accelerator group (shortcut keys) */
  accel_group = gtk_accel_group_new ();

  /* Make an ItemFactory (that makes a menubar) */
  item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>",
                                       accel_group);
  /* This function generates the menu items. Pass the item factory,
     the number of items in the array, the array itself, and any
     callback data for the the menu items. */
  gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);

  /* Attach the new accelerator group to the window. */
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

  /* Finally, return the actual menu bar created by the item factory. */
  menubar = gtk_item_factory_get_widget (item_factory, "<main>");
  gtk_box_pack_start (GTK_BOX (overall_box), menubar, FALSE, FALSE, 2);

  if (do_viewer)
  {
    /*****************************  the text box  ***************************/
    /* create scrollable window for the text box */
    text_box = gtk_scrolled_window_new (FALSE, FALSE);
    gtk_container_add (GTK_CONTAINER(overall_box), text_box);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (text_box),
                                    GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);
    /* set an inside border width of the window */
    gtk_container_set_border_width (GTK_CONTAINER (text_box), 5);
    /* create the TextView widget */
    viewer = gtk_text_view_new ();
    gtk_container_add (GTK_CONTAINER(text_box), viewer);
    /* obtain the text buffer object from the text view object */
    buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (viewer));
    g_signal_connect (G_OBJECT (buffer), "end-user-action",
                G_CALLBACK (viewer_callback), NULL);
    /* set monospace font */
    font = pango_font_description_from_string (TERMINAL_FONT);
    gtk_widget_modify_font (viewer, font);
  
    /* set black background */
//    gdk_color_parse ("black", &color);
//    gtk_widget_modify_base (viewer, GTK_STATE_NORMAL, &color);
    gtk_widget_set_size_request (window, 600, 460);
  } else
  {
    /***********************  the background image  *************************/
    backgnd = gdk_pixbuf_new_from_inline (-1, background_inline, FALSE, NULL);
    bg_image = gtk_image_new_from_pixbuf (backgnd);
    gtk_container_add (GTK_CONTAINER (overall_box), bg_image);
    gtk_widget_set_size_request (window,
      gdk_pixbuf_get_width (backgnd), gdk_pixbuf_get_height (backgnd)+120);
  }
  /* adjust the main window size to the background picture */
  /* start a timer for screen update */
  timer = gtk_timeout_add (20, timer_callback, NULL);
  
  /**************************  the status line  ***************************/
  statusline = gtk_label_new ("Waiting for DIY Calculator ... ");
  gtk_box_pack_start (GTK_BOX (overall_box), statusline, FALSE, FALSE, 10);
  gtk_misc_set_alignment (GTK_MISC(statusline), 0, 0);

  /***************************  the buttons  *****************************/
  /* create a box for the buttons */
  button_box = gtk_hbox_new (TRUE, 20);

  /* put into the window */
  gtk_box_pack_start (GTK_BOX (overall_box), button_box, FALSE, TRUE, 0);

  /* `Load' button */
  button = gtk_button_new_with_label ("Load");
  g_signal_connect (G_OBJECT(button),"clicked",G_CALLBACK(load_ram),NULL);
  gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

  /* `Reset' button */
  button = gtk_button_new_with_label ("Reset");
  g_signal_connect (G_OBJECT(button),"clicked",G_CALLBACK(do_reset),NULL);
  gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

  /* `Step' button */
  button = gtk_button_new_with_label ("Step");
  g_signal_connect (G_OBJECT(button),"clicked",G_CALLBACK(do_step),NULL);
  gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

  /* `Run' button */
  button = gtk_button_new_with_label ("Run");
  g_signal_connect (G_OBJECT(button),"clicked",G_CALLBACK(do_run),NULL);
  gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

  /* `Quit' button */
  button = gtk_button_new_with_label ("Quit");
  g_signal_connect (G_OBJECT(button),"clicked",G_CALLBACK(exit_program),NULL);
  gtk_box_pack_start (GTK_BOX(button_box), button, TRUE, TRUE, 0);

  /* show everything */
  gtk_widget_show_all (window);
}

/* init_port */
/**************************************************************************/
int init_port (              /* wrapper for open_port()                   */
  char *name)                /* which port                                */
/***************************************************************************
* globals     : portfd, offline, do_timer_cb, main_window                  *
* called from : main(), set_comm_port_close()                              *
* calls       : open_port(), set_mode(), do_connect(), gtk_msg_box()       *
*               send_cmd(), read_line()                                    *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* This is a wrapper for open_port() with error msg, etc.                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the port file descriptor or -1                                           *
***************************************************************************/
{
  char buf[BUFLEN];

  if (offline)
  {
    portfd = 0;
    set_mode (M_OFFLINE);
  } else
  {
    /* open the communications port */
    portfd = open_port (name);
    if (portfd == -1)
    {
      DEBUG (("ERROR: open_port(%s)\n", name));
      sprintf (buf, "Cannot open \"%s\"!", name);
      gtk_msg_box (main_window, "ERROR", buf, 0);
      set_mode (M_DISCONNECTED);
    } else
    {
      do_connect ();
      do_timer_cb = 0;
      send_cmd (RESET);
      /* eat response + prompt */
      read_line (buf, BUFLEN);
      read_line (buf, BUFLEN);
      read_line (buf, BUFLEN);
      do_timer_cb = 1;
    }
  }
  return portfd;
}

/* atox */
/**************************************************************************/
int atox (char *s)           /* convert string to HEX number              */
/***************************************************************************
* globals     :                                                            *
* called from : various                                                    *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Convert string to HEX number                                             *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* number (as integer)                                                      *
***************************************************************************/
{
  int digit, number;

  digit = 0;
  number = 0;
  while (isxdigit(*s))
  {
    if ('0' <= *s && *s <= '9')
      digit = *s - '0';
    else if ('A' <= *s && *s <= 'F')
      digit = *s - ('A' - 10);
    else if ('a' <= *s && *s <= 'f')
      digit = *s - ('a' - 10);
    number = (number << 4) + digit;
    s ++;
  }
  return number;
}

/* add_window_title */
/**************************************************************************/
void add_window_title (      /* add a string to the active window title   */
  GtkWidget *window,         /* the window                                */
  char *s)                   /* the string                                */
/***************************************************************************
* globals     : offline                                                    *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Add a string to the active window title.                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  static char window_title[BUFLEN];
  char   buf[BUFLEN];
  static int first=1;

  if (first)
  {
    first = 0;
    sprintf (window_title, "%s V%s %s %s  ",
      PROGNAME, VER, __DATE__, offline?" - offline ":"");
  }
  strcpy (buf, window_title);
  strcat (buf, s);
  gtk_window_set_title (GTK_WINDOW (window), buf);
}

/* usage */
/**************************************************************************/
void usage ()                /* print usage inforamtion and exit          */
/***************************************************************************
* globals     :                                                            *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Print usage information and exit.                                        *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
#ifdef WIN32
  int fdout;
  FILE *out;
  
  /* create a console window for stdio output */
  AllocConsole ();
  fdout = _open_osfhandle ((long)GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
  out = fdopen (fdout, "w");
  *stdout = *out;
#endif

  printf ("\n%s V%s  %s\n\n", PROGNAME, VER, __DATE__);
  puts("Usage:  diy_load [options]");
  puts("   [-d|--debug] .............. print debug information");
puts("   [-o|--offline] ............ offline mode (no DIY Calculator connected)");
  printf (
    "   [-p|--port=portname] ...... communications port name, default is %s\n",
    DEFAULT_PORT);
  puts("   [-t|--terminal] ........... terminal mode");
  puts("   [-v|--version] ............ print version information and exit");
  puts("\n   any other option brings this screen\n");

#ifdef WIN32
#ifdef RELEASE_VERSION
  printf ("\npress any key ... ");
  getch();
#endif
  FreeConsole ();
#endif
  
  exit (EXIT_SUCCESS);
}

/* printf_statusline */
/**************************************************************************/
void printf_statusline (     /* print formatted text to the status line   */
  char *fmt,                 /* format string                             */
  ...)                       /* arguments                                 */
/***************************************************************************
* globals     : statusline                                                 *
* called from : load_ram()                                                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Print one line of text to the status line.                               *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
  va_list  ap;
  char     buf[BUFLEN];

  va_start (ap, fmt);
  vsnprintf (buf, BUFLEN-1, fmt, ap);
  va_end (ap);

  gtk_label_set_text (GTK_LABEL(statusline), buf);
}

/* set_mode */
/**************************************************************************/
void set_mode (              /* set a new mode and update status line     */
  int new_mode)              /* the new mode: (dis-) connected, step, etc.*/
/***************************************************************************
* globals     : statusline, mode, mode_names[]                             *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Set a new global mode (connected, disconnected, single-step, run, etc.)  *
* and update the status line.                                              *
*                                                                          *
* Returns                                                                  *
* --------                                                                 *
* void                                                                     *
***************************************************************************/
{
  mode = new_mode;
  gtk_label_set_text (GTK_LABEL(statusline), mode_names[new_mode]);
}
