/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/* 
 * Filename: save_ram.c
 *
 * Function: the Save RAM dialog box
 *
 * Part of:  DIY Calculator loader application
 *
 *
 * Changelog
 *
 *  1.0  11.Nov.2006   JH   new
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>

#include <string.h>          /* strchr() */
#include <stdlib.h>          /* atoi() */

#ifndef WIN32
#include <unistd.h>          /* usleep() */
#endif

#include "diy_load.h"
#include "serial.h"
#include "cmd.h"
#include "cpu_clk.h"

#include "gtk_msgbox.h"
#include "gtk_statusbox.h"
#include "gtk_helper.h"      /* gtk_sleep() */

/******************************  Defines  *********************************/


/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/**************************  Private Functions  ***************************/


/************************  Function Prototypes  ***************************/


/******************************  Globals  *********************************/
static GtkWidget *window;    /* the CPU speed adjustment window           */
static GtkWidget *start_entry, *end_entry;

extern int debug;
extern int offline;
extern int do_timer_cb;

extern int start_addr;
extern int end_addr;

/* save_ram_window_destroy */
/**************************************************************************/
void save_ram_window_destroy (    /* destroy the Save RAM window          */
  GtkWidget *widget,              /* widget which triggered the callback  */
  gpointer data)                  /* user data                            */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the Save RAM window                             *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the Save RAM window.                                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* start_entry_activate_cb */
/**************************************************************************/
void start_entry_activate_cb(/* start_entry "activate" callback function  */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : end_entry                                                  *
* called from : pressing the ENTER key in the start_entry widget           *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* start_entry "activate" callback function                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_grab_focus (end_entry);
}

/* save_ram_button_cb */
/**************************************************************************/
void save_ram_button_cb (    /* Save/Close button callback function       */
  GtkWidget *widget,         /* widget which triggered the callback       */
  gpointer data)             /* user data                                 */
/***************************************************************************
* globals     : window, start_entry, end_entry, start_addr, end_addr,      *
*               offline, do_timer_cb                                       *
* called from : on clicking any of the Save/Close buttons                  *
* calls       : save_ram_window_destroy(), read_port(), write_port(),atox()*
*               gtk_status_box(), gtk_status_box_set_value(), gtk_sleep()  *
*               gtk_status_box_destroy(), gtk_msg_box()                    *
* Function                                                                 *
* --------                                                                 *
* Save/Close button callback function                                      *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
#define MAXTIME 5300
#define MAXSIZE 0x10000
#define DELAY   10
  char cmd[60], buf[BUFLEN];
  char *start_buf, *end_buf;
  int range;
  GtkWidget *statusbox;
  int i, timer, endtime, value;

  DEBUG (("save_ram_button_cb() "));

  if (data)
  {
    DEBUG (("Save button\n"));
    start_buf = (char *)gtk_entry_get_text (GTK_ENTRY(start_entry));
    start_addr = atox (start_buf);
    end_buf = (char *)gtk_entry_get_text (GTK_ENTRY(end_entry));
    end_addr = atox (end_buf);
    sprintf (buf, "Saving address $%4.4X to $%4.4X to Flash ... ",
      start_addr, end_addr);
    if (offline)
      statusbox = gtk_status_box (window, "Save to Flash - offline", buf,1);
    else
      statusbox = gtk_status_box (window, "Save to Flash", buf, 1);
    gtk_status_box_set_value (statusbox, 0);
    range = (end_addr - start_addr) & 0xffff;
    sprintf (cmd, "cpflash %s %x\n", start_buf, range);
    if (!offline)
    {
      DEBUG (("Sending cmd: %s", cmd));
      write_port (cmd, strlen(cmd));
    }
  
    /* wait for the prompt */
    endtime = (MAXTIME * range) / (MAXSIZE * DELAY);
//    DEBUG (("range=%d, endtime=%d\n", range, endtime));
    value = 0;
    do_timer_cb = 0;
    timer = 0;
    while (timer < endtime + 100)
    {
      if (offline)
      {
        if (timer == endtime - 2)
          strcpy (buf, "# ");
        else
          buf[0] = '\0';
      } else
      {
        i = read_port (buf, BUFLEN);
        if (i == -1)
          break;
        buf[i] = '\0';
      }
      value = (100 * timer) / endtime;
      gtk_status_box_set_value (statusbox, value);
      gtk_sleep (DELAY);
//      DEBUG (("buf=\"%s\", timer=%d\n", buf, timer));
      if (strstr (buf, "# ") != NULL)
        break;
      timer ++;
    }
//    DEBUG (("last timer=%d, last value=%d\n", timer, value));
    do_timer_cb = 1;
    gtk_status_box_destroy (statusbox);
    if (timer == endtime)
    {
      /* timeout */
      gtk_msg_box (window, "Save to Flash", "Timeout Error!", 0);
    } else
    {
      /* OK */
      statusbox = gtk_status_box(window,"Save to Flash","Success!",1);
      gtk_status_box_set_value (statusbox, 100);
      gtk_sleep (1000);
      gtk_status_box_destroy (statusbox);
    }
  } else
  {
    DEBUG (("Cancel button\n"));
  }
  /* close the window */
  save_ram_window_destroy (window, NULL);
}

/* do_save_ram */
/**************************************************************************/
GtkWidget *do_save_ram (     /* the Save RAM dialog box                   */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : window, offline, start_entry, end_entry,start_addr,end_addr*
* called from : save_ram()                                                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Create the Save RAM dialog box window.                                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* pointer to the Save RAM window object                                    *
***************************************************************************/
{
  GtkWidget *overall_box, *hbox1, *hbox2;
  GtkWidget *label, *button;
  char buf[BUFLEN];

  DEBUG (("do_save_ram()\n"));

  /*****************  create the "Save to Flash" window  ******************/
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  /* install the delete_event handler that immediately exits GTK */
  g_signal_connect (G_OBJECT(window), "delete_event",
    G_CALLBACK (save_ram_window_destroy), NULL);
  /* modal window */
  gtk_window_set_modal (GTK_WINDOW(window), TRUE);
  /* set the border width of the window */
  gtk_container_set_border_width (GTK_CONTAINER (window), 10);
  /* set the window title */
  if (offline)
    gtk_window_set_title (GTK_WINDOW (window), "Save to Flash - offline");
  else
    gtk_window_set_title (GTK_WINDOW (window), "Save to Flash");
  /* set it on top of the main window */
  gtk_window_set_transient_for (GTK_WINDOW(window), GTK_WINDOW(parent));

  /*******************   create the various widgets  **********************/
  overall_box = gtk_vbox_new (FALSE, 10);
  /* put the box into the window */
  gtk_container_add (GTK_CONTAINER (window), overall_box);

  /* a hbox for the labels and entry boxes */
  hbox1 = gtk_hbox_new (FALSE, 10);
  gtk_container_add (GTK_CONTAINER (overall_box), hbox1);

  label = gtk_label_new ("From addr: $");
  gtk_box_pack_start (GTK_BOX(hbox1), label, FALSE, TRUE, 0);

  start_entry = gtk_entry_new ();
  sprintf (buf, "%4.4X", start_addr);
  gtk_entry_set_text (GTK_ENTRY(start_entry), buf);
  g_signal_connect (G_OBJECT (start_entry), "activate",
    G_CALLBACK (start_entry_activate_cb), NULL);
  gtk_widget_set_size_request (GTK_WIDGET (start_entry), 50, -1);
  gtk_container_add (GTK_CONTAINER (hbox1), start_entry);

  label = gtk_label_new ("to addr: $");
  gtk_box_pack_start (GTK_BOX(hbox1), label, FALSE, TRUE, 0);

  end_entry = gtk_entry_new ();
  sprintf (buf, "%4.4X", end_addr);
  gtk_entry_set_text (GTK_ENTRY(end_entry), buf);
  gtk_widget_set_size_request (GTK_WIDGET (end_entry), 50, -1);
  g_signal_connect (G_OBJECT (end_entry), "activate",
    G_CALLBACK (save_ram_button_cb), (gpointer)1);
  gtk_container_add (GTK_CONTAINER (hbox1), end_entry);

  /* the buttons */
  hbox2 = gtk_hbox_new (TRUE, 20);
  gtk_container_add (GTK_CONTAINER (overall_box), hbox2);

  button = gtk_button_new_with_label ("Save");
  g_signal_connect (G_OBJECT (button), "clicked",
    G_CALLBACK (save_ram_button_cb), (gpointer)1);
  gtk_box_pack_start (GTK_BOX(hbox2), button, TRUE, TRUE, 0);
  button = gtk_button_new_with_label ("Cancel");
  g_signal_connect (G_OBJECT (button), "clicked",
    G_CALLBACK (save_ram_button_cb), (gpointer)0);
  gtk_box_pack_start (GTK_BOX(hbox2), button, TRUE, TRUE, 0);

  gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_CENTER_ON_PARENT);

  /* show the window */
  gtk_widget_show_all (window);

  return window;
}
