/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: serial.c
 *
 * Function: serial I/O module
 *
 * Part of:  DIY Calculator loader application
 *
 * Changelog
 *
 *  0.1  13.Oct.2006   JH   new
 *  0.2  24.Oct.2006   JH   new design
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */

#include <gtk/gtk.h>

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>          /* access(), R_OK, W_OK, F_OK */
#include <fcntl.h>           /* O_RDWR, O_NOCTTY */
#include <termios.h>         /* struct termios */
#endif

#include <string.h>          /* strcpy(), strcasecmp() */
#include <ctype.h>           /* isdigit() */
#include <errno.h>

#include "serial.h"          /* struct COMM_PORT */

/******************************  Defines  *********************************/
#define MAX_VALUE_STRING 200
#define MAX_VALUE_NAME   100


/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/****************************  Private Functions  *************************/


/******************************  Globals  *********************************/
extern struct COMM_PORT portlist[];

#ifdef __linux__
static struct termios old_settings;    /* for restoring original settings */
#endif

extern int debug;
extern int portfd;

/* get_avail_ports */
/**************************************************************************/
int get_avail_ports ()       /* find the available communication ports    */
/***************************************************************************
* globals     : portlist[]                                                 *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* find the available communication ports                                   *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of communication ports                                            *
***************************************************************************/
{
#ifdef WIN32
  int   found;
  HKEY  hkey;
  char  value_string[MAX_VALUE_STRING];
  DWORD value_string_len;
  char  value_name[MAX_VALUE_NAME];
  DWORD value_name_len;
  DWORD type_code;
  DWORD res;
  int   i, j, k, number;
  char  *p;

  DEBUG (("get_avail_ports():\n"));
  found = 0;
  k = 0;
  if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, "HARDWARE\\DEVICEMAP\\SERIALCOMM",
                   0, KEY_READ, &hkey) == ERROR_SUCCESS)
  {
    /* find all key values */
    i = 0;
    do
    {
      value_name_len = sizeof (value_name);
      value_name[0] = '\0';
      value_string_len = sizeof (value_string);
      value_string[0] = '\0';
      res = RegEnumValue (
              hkey,               // key handle (in)
              i,                  // index of the value to be retrieved (in)
              value_name,         // buffer for value name (out)
              &value_name_len,    // size of value name (in/out)
              NULL,               // reserved
              &type_code,         // type code (out)
              value_string,       // buffer for value contents (out)
              &value_string_len); // size of value contents (in/out)
      if (res == ERROR_SUCCESS)
      {
        /* insert the found port to the list, sorted */
        for (p=value_string; !isdigit(*p) && *p; p++)
          ;
        number = atoi(p);
        for (j=0; j<found; j++)
        {
          if (number < portlist[j].number)
          {
            /* move already detected ports */
            for (k=found; k>j; k--)
            {
              strcpy (portlist[k].name, portlist[k-1].name);
              portlist[k].number = portlist[k-1].number;
              portlist[k].is_virtual = portlist[k-1].is_virtual;
            }
          } else
            k = found;
        }
        /* insert the newly found string */
        DEBUG (("  %s", value_string));
        strcpy (portlist[k].name, value_string);
        portlist[k].number = number;
        portlist[k].is_virtual = 0;
        if (strstr (value_name, "Device") == NULL)
        {
          /* is not a real hardware port */
          DEBUG ((" virtual"));
          portlist[k].is_virtual = 1;
        }
        DEBUG (("\n"));
        found ++;
      }
      i ++;
    } while (res == ERROR_SUCCESS);
  }
  DEBUG (("\n"));
  return found;
#else  /* WIN32 */
  int x, found;
  char portname[20];

  DEBUG (("get_avail_ports():\n"));
  for (x=0; x<MAX_COMM_PORTS; x++)
    portlist[x].number = -1;

  found = 0;

  /* search the "/dev/ttySxx" ports */
  for (x=0; x<=99; x++)
  {
    sprintf (portname, "/dev/ttyS%d", x);
    if (access(portname, R_OK|W_OK) == 0)
    {
      DEBUG (("  %s\n", portname));
      strcpy(portlist[found].name, portname);
      portlist[found].number = x;
      found ++;
    }
  }
  /* search the "/dev/ttyxx" ports */
  for (x=0; x<=99; x++)
  {
    sprintf (portname, "/dev/vpty%d", x);
    if (access(portname, F_OK) == 0)
    {
      sprintf (portname, "/dev/tty%d", x);
      if (access(portname, R_OK|W_OK) == 0)
      {
        DEBUG (("  %s, virtual\n", portname));
        strcpy (portlist[found].name, portname);
        portlist[found].number = x;
        portlist[found].is_virtual = 1;
        found ++;
      }
    }
  }
  DEBUG (("\n"));
  return found;
#endif /* Linux */
}

/* open_port */
/**************************************************************************/
int open_port (              /* open a specified communcations port       */
  char *portname)            /* which port to open                        */
/***************************************************************************
* globals     : portlist[], old_settings                                   *
* called from : init_port()                                                *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Open a specified communcations port.                                     *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* "file descriptor" for port                                               *
* -1 .. error                                                              *
***************************************************************************/
{
#ifdef WIN32
  int i;
  HANDLE hport;
  COMMCONFIG settings;
  DCB dcb;
  COMMTIMEOUTS ct;

  /* search the desired port inside the port list */
  for (i=0; i<MAX_COMM_PORTS; i++)
  {
    if (strcasecmp(portlist[i].name, portname) == 0)
      break;
  }
  if (portlist[i].number == 0)
  {
    /* not found => error */
    DEBUG (("open_port(): could not find \"%s\" in port list\n", portname));
    errno = ENOENT;
    return -1;
  }
  
  /* work around the "COM10" issue */
  if (portlist[i].number >= 10)
    sprintf (portname, "\\\\.\\%s", portlist[i].name);

  /* open the port */
  DEBUG (("open_port(): opening \"%s\" ... ", portlist[i].name));
  hport = CreateFile (
            portname,
            GENERIC_READ | GENERIC_WRITE,
            0,
            0,
            OPEN_EXISTING,
            0,
            0
          );
  if (hport == INVALID_HANDLE_VALUE)
  {
    DEBUG (("ERROR\n"));
    errno = EACCES;
    return -1;
  }
  /* set the communications parameters */
  /* the DCB struct */
  dcb.DCBlength = sizeof(DCB);
  dcb.BaudRate = CBR_38400;
  dcb.fBinary = TRUE;
  dcb.fParity = FALSE;
  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fDtrControl = DTR_CONTROL_ENABLE;
  dcb.fDsrSensitivity = FALSE;
  dcb.fTXContinueOnXoff = TRUE;
  dcb.fOutX = FALSE;
  dcb.fInX = FALSE;
  dcb.fErrorChar = FALSE;
  dcb.fNull = FALSE;
  dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
  dcb.fAbortOnError = FALSE;
  dcb.wReserved = 0;
  dcb.XonLim = 16;
  dcb.XoffLim = 4;
  dcb.ByteSize = 8;
  dcb.Parity = NOPARITY;
  dcb.StopBits = TWOSTOPBITS;
  dcb.XonChar = '\021';   /* CTRL Q */
  dcb.XoffChar = '\023';  /* CTRL S */
  dcb.ErrorChar = '\007'; /* ring the bell on error */
  dcb.EofChar = '\004';   /* CTRL D */
  dcb.EvtChar = '\000';   /* ?? */

  /* the COMMCONFIG struct */
  settings.dwSize = sizeof(COMMCONFIG);
  settings.wVersion = 1;  /* ?? */
  /* settings.wReserved ... do not use */
  memcpy (&settings.dcb, &dcb, sizeof(DCB));
  settings.dwProviderSubType = PST_RS232;
  settings.dwProviderOffset = 0;
  settings.dwProviderSize = 0;

  if (SetCommConfig(hport, &settings, sizeof(COMMCONFIG)) == 0)
  {
    DEBUG (("cannot set configuration\n"));
    errno = EACCES;
    CloseHandle (hport);
    return -1;
  }
  
  /* dont use any CommTimeouts */
  ct.ReadIntervalTimeout = MAXDWORD;
  ct.ReadTotalTimeoutMultiplier = 0;
  ct.ReadTotalTimeoutConstant = 0;
  ct.WriteTotalTimeoutMultiplier = 0;
  ct.WriteTotalTimeoutConstant = 0;
  if (SetCommTimeouts (hport, &ct) == 0)
  {
    DEBUG (("cannot set comm timeouts\n"));
    errno = EACCES;
    CloseHandle (hport);
    return -1;
  }

  DEBUG (("OK, hport=0x%x\n", (unsigned)hport));
  return (int)hport;
#else  /* WIN32 */
  int i;
  int fd;
  struct termios settings;

  /* search the desired port inside the port list */
  for (i=0; i<MAX_COMM_PORTS; i++)
  {
    if (strcasecmp(portlist[i].name, portname) == 0)
      break;
  }
  if (portlist[i].number == -1)
  {
    /* not found => error */
    DEBUG (("open_port(): could not find \"%s\" in port list\n", portname));
    errno = ENOENT;
    return -1;
  }
  DEBUG (("open_port(): opening \"%s\" ... ", portlist[i].name));
  fd = open (portlist[i].name, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd == -1)
  {
    DEBUG (("ERROR\n"));
    return -1;
  }
  DEBUG (("OK, fd=%d\n", fd));

  /* set comm. parameters */
  tcgetattr (fd, &old_settings);
  settings.c_cflag = B38400 | CS8 | CREAD | CLOCAL | CSTOPB;
  settings.c_iflag = 0;
  settings.c_oflag = 0;
  settings.c_lflag = 0;

  /* comm. port setup is done */
  tcflush (fd, TCIOFLUSH);
  tcsetattr (fd, TCSANOW, &settings);

  return fd;
#endif /* Linux */
}

/* close_port */
/**************************************************************************/
int close_port (             /* close a specified communcations port      */
  int fd)                /* fd of port to close                       */
/***************************************************************************
* globals     : old_settings                                               *
* called from : main()                                                     *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Close a specified communcations port.                                    *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* 0 .. OK                                                                  *
* -1 .. error                                                              *
***************************************************************************/
{
#ifdef WIN32
  HANDLE hport;
  DWORD res;

  hport = (HANDLE) fd;
  DEBUG (("close_port(): hport=0x%x ... ", (unsigned int)hport));
  if (hport == INVALID_HANDLE_VALUE)
  {
    errno = EBADF;
    DEBUG (("invalid handle\n"));
    return -1;
  }

  res = CloseHandle (hport);
  if (res != 0)
  {
    DEBUG(("OK\n"));
  } else
  {
    DEBUG(("ERROR\n"));
    errno = EIO;
    return -1;
  }
  return 0;
#else  /* WIN32 */
  int res;

  DEBUG (("close_port(): fd=%d ... ", fd));

  /* restore original port settings */
  if (tcsetattr (fd, TCSANOW, &old_settings))
  {
    DEBUG (("ERROR restoring port settings\nclosing port ... "));
  }

  res = close (fd);
  if (res == 0)
  {
    DEBUG(("OK\n"));
  } else
  {
    DEBUG(("ERROR\n"));
  }
  return res;
#endif /* Linux */
}

/* write_port */
/**************************************************************************/
int write_port (             /* write to the communications port          */
  void *buf,                 /* the buffer                                */
  size_t count)              /* length of the buffer                      */
/***************************************************************************
* globals     : portfd                                                     *
* called from :                                                            *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Write a buffer to the communications port.                               *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of bytes written .. OK                                            *
* -1 .. error                                                              *
***************************************************************************/
{
#ifdef WIN32
  HANDLE hport;
  DWORD bytes_written;
  int res;

  hport = (HANDLE) portfd;
  res = WriteFile (hport, buf, count, &bytes_written, NULL);
  if (res == 0)
  {
    switch (GetLastError())
    {
      case ERROR_INVALID_USER_BUFFER:
        errno = EFAULT;
        break;
      case ERROR_NOT_ENOUGH_MEMORY:
        errno = ENOSPC;
        break;
    }
    return -1;
  }
  return bytes_written;
#else  /* WIN32 */
  return write (portfd, buf, count);
#endif /* Linux */
}

/* read_port */
/**************************************************************************/
int read_port (              /* read a buffer from a communications port  */
  void *buf,                 /* the buffer                                */
  size_t count)              /* length of the buffer                      */
/***************************************************************************
* globals     : portfd                                                     *
* called from : read_line(), wait_for_string(), timer_callback(),          *
*               save_ram_button_cb()                                       *
* calls       : char_avail()                                               *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Read a buffer from a communications port.                                *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of bytes read .. OK                                               *
* -1 .. error                                                              *
***************************************************************************/
{
#ifdef WIN32
  HANDLE hport;
  DWORD bytes_read;
  int i, res;
  char *cbuf;
  
  hport = (HANDLE) portfd;
  cbuf = (char *) buf;
  for (i=0; i<count; i++)
  {
    res = ReadFile (hport, &cbuf[i], 1, &bytes_read, NULL);
    if (res == 0)
    {
      switch (GetLastError())
      {
        case ERROR_INVALID_USER_BUFFER:
          errno = EFAULT;
          break;
        case ERROR_NOT_ENOUGH_MEMORY:
          errno = ENOSPC;
          break;
      }
      cbuf[i] = '\0';
      return -1;
    }
    if (bytes_read == 0)
      break;
  }
  return i;
#else
  int i, res;
  char *cbuf;
  
  cbuf = (char *) buf;
  for (i=0; i<count; i++)
  {
    if (!char_avail (portfd))
      break;
    res = read (portfd, &cbuf[i], 1);
    if (res == -1)
    {
      cbuf[i] = '\0';
      return -1;
    }
  }
  return i;
#endif
}

/* read_line */
/**************************************************************************/
int read_line (              /* read a text line from a comm. port        */
  void *buf,                 /* the buffer                                */
  size_t count)              /* max. length of the buffer                 */
/***************************************************************************
* globals     :                                                            *
* called from : various                                                    *
* calls       : read_port()                                                *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Read a text line from a communications port.                             *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of bytes read .. OK                                               *
* -1 .. error                                                              *
***************************************************************************/
{
  int i, res;
  char *cbuf;

//  DEBUG (("readline()\n"));
  cbuf = (char *) buf;
  count --;
  for (i=0; i<count; i++)
  {
    res = read_port (&cbuf[i], 1);
    if (res == -1)
    {
      cbuf[i] = '\0';
      return -1;
    }
    if (res == 0)
      break;
    if (cbuf[i] == '\n')
      break;
  }
  cbuf[i+1] = '\0';
//  DEBUG (("readline(): buf=\"%s\", i=%d, res=%d\n", cbuf, i, res));
  return i;
}

#ifndef WIN32
/* char_avail */
/**************************************************************************/
int char_avail (             /* check for something available             */
  int fd)                /* the communications port file descriptor   */
/***************************************************************************
* globals     :                                                            *
* called from : read_port                                                  *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Check if something is available from the communications port.            *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* number of characters available or 0                                      *
***************************************************************************/
{
  fd_set rd;
  struct timeval tv;

//  DEBUG (("char_avail(): "));

  tv.tv_sec = 0;
  tv.tv_usec = 0;
  FD_ZERO (&rd);
  FD_SET (fd, &rd);
  select (fd+1, &rd, NULL, NULL, &tv);
//  DEBUG (("select() returns: %d\n", res));
  return FD_ISSET (fd, &rd);
}
#endif
