/* 
 * Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
/* 
 * Filename: version_info.c
 *
 * Function: display version information
 *
 * Part of:  DIY Calculator loader application
 *
 * Changelog
 *
 *  0.1   6.Nov.2006   JH   new
 *  1.0  11.Nov.2006   JH   first delivered version
 *  1.1  14.Nov.2006   JH   code review
 *
 */


#include <gtk/gtk.h>

#ifdef WIN32
#include <windows.h>         /* Sleep() */
#else
#include <unistd.h>          /* usleep() */
#endif

#include <string.h>

#include "diy_load.h"
#include "cmd.h"             /* send_cmd() */
#include "serial.h"          /* read_line() */
#include "cpu_clk.h"         /* divider_to_freq() */


/******************************  Defines  *********************************/


/******************************  Macros  **********************************/
#define DEBUG(x)  if (debug) g_print x  /* a simple trace macro */

/******************************  Structs  *********************************/


/**************************  Private Functions  ***************************/

/******************************  Globals  *********************************/
static GtkWidget *window;

extern int debug;
extern int offline;
extern int do_timer_cb;


/* version_window_destroy */
/**************************************************************************/
void version_window_destroy (     /* destroy the version display window   */
  GtkWidget *widget,              /* widget which triggered the callback  */
  gpointer data)                  /* user data                            */
/***************************************************************************
* globals     : window                                                     *
* called from : on closing the version info display window                 *
* calls       :                                                            *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Destroy the version info display window.                                 *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* void                                                                     *
***************************************************************************/
{
  gtk_widget_destroy (window);
  window = NULL;
}

/* do_version */
/**************************************************************************/
GtkWidget *do_version_info ( /* display the version info window           */
  GtkWidget *parent)         /* the parent window                         */
/***************************************************************************
* globals     : window, offline, do_timer_cb                               *
* called from : version_info()                                             *
* calls       : send_cmd(), read_line(), divider_to_freq()                 *
*                                                                          *
* Function                                                                 *
* --------                                                                 *
* Display the version info window.                                         *
*                                                                          *
* Returns                                                                  *
* -------                                                                  *
* the version info window                                                  *
***************************************************************************/
{
  GtkWidget *vbox, *table, *label, *button, *entry;
  char buf[BUFLEN], *p;
  unsigned char divider;
  unsigned int freq;

  DEBUG (("do_version_info()\n"));

  do_timer_cb = 0;
  if (!offline)
  {
    send_cmd (INFO);
    usleep (200000);
  
    /* eat empty lines */
    read_line (buf, BUFLEN);
    read_line (buf, BUFLEN);
    read_line (buf, BUFLEN);
  }

  /* create a new top level window */
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_modal (GTK_WINDOW(window), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(parent));
  gtk_window_set_title (GTK_WINDOW (window), "Version Information");
  gtk_container_set_border_width (GTK_CONTAINER (window), 10);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  table = gtk_table_new (6, 2, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE(table), 10);
  gtk_table_set_col_spacings (GTK_TABLE(table), 10);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);

  label = gtk_label_new ("FPGA version:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), "*****");
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 0, 1);
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(entry), "offline");
  else
  {
    read_line (buf, BUFLEN);
    DEBUG (("do_version_info() line 1: %s", buf));
    p = strchr (buf, '$');
    if (p)
    {
      buf[0] = p[1];
      buf[1] = '.';
      buf[2] = p[2];
      buf[3] = '\0';
      gtk_entry_set_text (GTK_ENTRY(entry), buf);
    }
  }

  label = gtk_label_new ("CPU version:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), "*****");
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 1, 2);
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(entry), "offline");
  else
  {
    read_line (buf, BUFLEN);
    DEBUG (("do_version_info() line 2: %s", buf));
    p = strchr (buf, '$');
    if (p)
    {
      buf[0] = p[1];
      buf[1] = '.';
      buf[2] = p[2];
      buf[3] = '\0';
      gtk_entry_set_text (GTK_ENTRY(entry), buf);
    }
  }

  label = gtk_label_new ("ROM monitor version:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), "*****");
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 2, 3);
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(entry), "offline");
  else
  {
    read_line (buf, BUFLEN);
    DEBUG (("do_version_info() line 3: %s", buf));
    p = strchr (buf, 'V');
    if (p)
    {
      p ++;
      p[3] = '\0';
      gtk_entry_set_text (GTK_ENTRY(entry), p);
    }
  }

  label = gtk_label_new ("CPU clock speed:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 3, 4);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), "*****");
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 3, 4);
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(entry), "offline");
  else
  {
    read_line (buf, BUFLEN);
    DEBUG (("do_version_info() line 4: %s", buf));
    p = strrchr (buf, '$');
    if (p)
    {
      p ++;
      p[2] = '\0';
      divider = (unsigned char) atox(p);
      freq = divider_to_freq (divider, buf);
      gtk_entry_set_text (GTK_ENTRY(entry), buf);
    }
  }

  label = gtk_label_new ("SPI flash signature:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 4, 5);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), "*****");
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 4, 5);
  if (offline)
    gtk_entry_set_text (GTK_ENTRY(entry), "offline");
  else
  {
    read_line (buf, BUFLEN);
    DEBUG (("do_version_info() line 5: %s", buf));
    p = strchr (buf, '$');
    if (p)
    {
      buf[0] = p[1];
      buf[1] = '.';
      buf[2] = p[2];
      buf[3] = '\0';
      gtk_entry_set_text (GTK_ENTRY(entry), buf);
    }
  }

  label = gtk_label_new ("Loader version:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 5, 6);
  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text (GTK_ENTRY(entry), VER);
  gtk_entry_set_alignment (GTK_ENTRY(entry), 0.5);
  gtk_widget_set_size_request (GTK_WIDGET(entry), 60, 20);
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 5, 6);

  /* Close button */
  button = gtk_button_new_with_label ("Close");
  g_signal_connect (G_OBJECT (button), "clicked",
    G_CALLBACK (version_window_destroy), NULL);
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, TRUE, 0);
  gtk_widget_grab_focus (button);

  /* put the dialog box to the center of the main window */
  gtk_window_set_position (GTK_WINDOW(window),GTK_WIN_POS_CENTER_ON_PARENT);

  gtk_widget_show_all (window);
  do_timer_cb = 1;

  return window;
}
